import asyncio
import collections
import difflib
import os
import re

import async_lru

from ... import constants, errors, utils
from . import SceneCheckResult

import logging  # isort:skip
_log = logging.getLogger(__name__)


class MultiPredbApi:
    """
    Wrapper around multiple :class:`~.PredbApiBase` instances

    Each method loops over every provided predb and calls the same method on
    it. The first call that does not raise :class:`~.RequestError` or
    :class:`NotImplementedError` is returned.

    If no call succeeds and any exceptions were raised, they are combined into a
    single :class:`~.RequestError`, which is raised. If every call raised
    :class:`NotImplementedError`, a combined :class:`NotImplementedError` is
    raised.

    :param predbs: Sequence of :class:`~.PredbApiBase` instances
    """

    DEFAULT_PREDB_NAMES = (
        'srrdb',
        # 'corruptnet',
        # 'predbovh',
        # 'predbclub',
        'predbde',
    )

    def __init__(self, predbs=None):
        if predbs:
            self._predbs = predbs
        else:
            self._predbs = [
                utils.predbs.predb(name)
                for name in type(self).DEFAULT_PREDB_NAMES
            ]

    @property
    def predbs(self):
        """Sequence of :class:`~.PredbApiBase` that are used"""
        return tuple(self._predbs)

    async def _call(self, predb, method_name, *args, **kwargs):
        method = getattr(predb, method_name)
        return await method(*args, **kwargs)

    def _raise(self, exceptions, method_name):
        if not exceptions:
            raise NotImplementedError(
                '|'.join(predb.name for predb in self._predbs)
                + f'.{method_name}'
            )
        elif len(exceptions) == 1:
            raise exceptions[0]
        else:
            raise errors.RequestError(
                'All queries failed: ' + ', '.join(str(e) for e in exceptions)
            )

    async def _for_each_predb(self, method_name, *args, **kwargs):
        # Try calling `method_name` on each predb sequentially and return the
        # first successful response.
        exceptions = []
        for predb in self._predbs:
            _log.debug('Trying %s.%s', predb.name, method_name)
            try:
                return await self._call(predb, method_name, *args, **kwargs)
            except (NotImplementedError, errors.RequestError) as e:
                _log.debug('Collecting exception: %r', e)
                exceptions.append(e)

        # All predbs are unreachable or don't have `method_name` implemented.
        self._raise(exceptions, method_name)

    async def _for_all_predbs(self, method_name, *args, **kwargs):
        # Call `method_name` on all predbs concurrently and yield the return
        # value or RequestError/NotImplementedError as they occur.
        method_calls = [
            asyncio.create_task(
                self._call(predb, method_name, *args, **kwargs)
            )
            for predb in self._predbs
        ]

        try:
            for coro in asyncio.as_completed(method_calls):
                try:
                    result = await coro
                except (NotImplementedError, errors.RequestError) as e:
                    yield e
                else:
                    yield result
        finally:
            # If our caller stops iterating over our yielded `result`s, we may
            # still have unfinished tasks that will continue to run in the
            # background. We cancel them here because
            #   1. nobody cares about their return values anymore
            #
            #   2. if one of the tasks raises an exception, asyncio will
            #      eventually complain about an unhandled exception. Cancelling
            #      the tasks seems to prevents that.
            #
            # NOTE: This is untested functionality because I couldn't reproduce
            #       this issue in a test case.
            for task in method_calls:
                task.cancel()

    @async_lru.alru_cache
    async def search(self, *args, **kwargs):
        """
        Search multiple predbs concurrently and return first non-empty results

        See :meth:`~.PredbApiBase.search`.
        """
        exceptions = []
        async for results in self._for_all_predbs('search', *args, **kwargs):
            if isinstance(results, Exception):
                exceptions.append(results)
            else:
                # Return non-empty search results or wait for next response
                if results:
                    return results

        if exceptions:
            # All predbs are unreachable or don't have `method_name` implemented.
            # Raise combined RequestErrors.
            self._raise(exceptions, method_name='search')
        else:
            # No search results
            return []

    @async_lru.alru_cache
    async def release_files(self, *args, **kwargs):
        """See :meth:`~.PredbApiBase.release_files`"""
        return await self._for_each_predb('release_files', *args, **kwargs)

    _nogroup_regexs = (
        re.compile(r'^$'),
        re.compile(r'^(?i:nogroup)$'),
        re.compile(r'^(?i:nogrp)$'),
    )

    @async_lru.alru_cache
    async def is_scene_release(self, release):
        """
        Return whether `release` is a scene release or not

        .. note:: A renamed or otherwise altered scene release is still
            considered a scene release.

        :param release: Release name, path to release

        :return: :class:`~.types.SceneCheckResult`
        """
        if isinstance(release, str):
            release_info = utils.release.ReleaseInfo(release)
        else:
            release_info = release

        # Empty group or names like "NOGROUP" are non-scene
        if (
            # Any NOGROUP-equivalent means it's not scene
            any(regex.search(release_info['group']) for regex in self._nogroup_regexs)

            # Abbreviated file names also have an empty group, but ReleaseInfo()
            # can pick up information from the parent directory and we can do a
            # successful search for that.
            and not utils.predbs.is_abbreviated_filename(release_info.path)
        ):
            return SceneCheckResult.false

        # We know at least `title` and `group`. If there are any search results,
        # it should be a scene release.
        results = await self.search(release)
        if results:
            return SceneCheckResult.true
        else:
            return SceneCheckResult.false

    async def verify_release_name(self, content_path, release_name):
        """
        Raise if release was renamed

        :param content_path: Path to release file or directory
        :param release_name: Known exact release name, e.g. from :meth:`search`
            results

        :raise SceneRenamedError: if release was renamed
        :raise SceneError: if `release_name` is not a scene release
        :raise NotImplementedError: if :meth:`release_files` raises it
        """
        content_path = content_path.strip(os.sep)
        _log.debug('Verifying release name: %r =? %r', content_path, release_name)

        # TODO: Rewrite this comment: We MUST first check if release_files() is implemented. If we check
        # is_scene_release() first and an incomplete predb wrongly reports "not
        # a scene release", MultiPredbApi can move on the next predb.
        files = await self.release_files(release_name)

        # TODO: Rewrite this comment: We check is_scene_release() a) because we can fail early before we do
        # all the work below and b) because we don't want to raise
        # SceneRenamedError below if this is not a scene release.
        if not await self.is_scene_release(release_name):
            raise errors.SceneError(f'Not a scene release: {release_name}')

        # Figure out which file is the actual payload. Note that the release may not
        # contain any files, e.g. Wrecked.2011.DiRFiX.LIMITED.FRENCH.720p.BluRay.X264-LOST.
        main_release_file = None
        if files:
            content_file_extension = utils.fs.file_extension(content_path)
            if content_file_extension:
                # `content_path` points to a file, not a directory
                content_file = utils.fs.basename(content_path)
                # Find the closest match in the released files
                # (`content_file` may have been renamed)
                filename_matches = difflib.get_close_matches(content_file, files)
                if filename_matches:
                    main_release_file = filename_matches[0]

            if not main_release_file and files:
                # Default to the largest file
                main_release_file = sorted(
                    (info for info in files.values()),
                    key=lambda info: info['size'],
                )[0]['file_name']

        # No files in this release, default to `release_name`
        if not main_release_file:
            main_release_file = release_name
        _log.debug('Main release file: %r', main_release_file)

        # Generate set of paths that are valid for this release
        acceptable_paths = {release_name}

        # Properly named directory that contains the released file. This covers
        # abbreviated files and all other files.
        for file in files:
            acceptable_paths.add(os.path.join(release_name, file))

        # Any non-abbreviated files may exist outside of a properly named parent
        # directory
        for file in files:
            if not utils.predbs.is_abbreviated_filename(file):
                acceptable_paths.add(file)

        # If `release_name` is an episode, it may be inside a season pack parent
        # directory. This only matters if we're dealing with an abbreviated file
        # name; normal file names are independent of their parent directory name.
        if utils.predbs.is_abbreviated_filename(content_path):
            season_pack_name = utils.predbs.common.get_season_pack_name(release_name)
            for file in files:
                acceptable_paths.add(f'{season_pack_name}/{file}')

        # Standalone file is also ok if it is named `release_name` with the same
        # file extension as the main file
        main_release_file_extension = utils.fs.file_extension(main_release_file)
        acceptable_paths.add('.'.join((release_name, main_release_file_extension)))

        # Release is correctly named if `content_path` ends with any acceptable path
        for path in (p.strip(os.sep) for p in acceptable_paths):
            if re.search(rf'(?:^|{re.escape(os.sep)}){re.escape(path)}$', content_path):
                return

        # All attempts to match `content_path` against `release_name` have failed.
        # Produce a useful error message.
        if utils.predbs.is_abbreviated_filename(content_path):
            # Abbreviated files should never be handled without a parent
            original_name = os.path.join(release_name, main_release_file)
        elif utils.fs.file_extension(content_path):
            # Assume `content_path` refers to a file, not a directory
            # NOTE: We can't use os.path.isdir(), `content_path` may not exist.
            original_name = main_release_file
        else:
            # Assume `content_path` refers to directory
            original_name = release_name

        # Use the same number of parent directories for original/existing path. If
        # original_name contains the parent directory, we also want the parent
        # directory in existing_name.
        original_name_parts_count = original_name.count(os.sep)
        content_path_parts = content_path.split(os.sep)
        existing_name = os.sep.join(content_path_parts[-original_name_parts_count - 1:])

        raise errors.SceneRenamedError(
            original_name=original_name,
            existing_name=existing_name,
        )

    async def verify_release_files(self, content_path, release_name):
        """
        Check if existing files have the correct size

        :param content_path: Path to release file or directory
        :param release_name: Known exact release name, e.g. from :meth:`search`
            results

        The return value is a sequence of :class:`~.errors.SceneError` exceptions:

            * :class:`~.errors.SceneFileSizeError` if a file has the wrong size
            * :class:`~.errors.SceneMissingInfoError` if the correct file size
              of file cannot be found
            * :class:`~.errors.SceneError` if `release_name` is not a scene release
        """
        exceptions = []

        # TODO: Rewrite this comment: We MUST first check if release_files() is implemented. If we check
        # is_scene_release() first and an incomplete predb wrongly reports "not
        # a scene release", MultiPredbApi can move on the next predb.
        fileinfos = await self.release_files(release_name)

        # TODO: Rewrite this comment: We check is_scene_release() a) because we can fail early before we do
        # all the work below and b) because we don't want to raise
        # SceneRenamedError below if this is not a scene release.
        if not await self.is_scene_release(release_name):
            exceptions.append(errors.SceneError(f'Not a scene release: {release_name}'))
        if exceptions:
            return tuple(exceptions)

        def get_release_filesize(filename):
            return fileinfos.get(filename, {}).get('size', None)

        # Map file paths to expected file sizes
        if os.path.isdir(content_path):
            # Map each file in content_path to its correct size
            exp_filesizes = {
                filepath: get_release_filesize(utils.fs.basename(filepath))
                for filepath in utils.fs.file_list(content_path)
            }
        else:
            # `content_path` is file
            if len(fileinfos) == 1:
                # Original release is also a single file, but it may be in a directory
                filename = tuple(fileinfos)[0]
                exp_filesize = get_release_filesize(filename)
                exp_filesizes = {
                    # Title.2015.720p.BluRay.x264-FOO.mkv
                    content_path: exp_filesize,
                    # Title.2015.720p.BluRay.x264-FOO/foo-title.mkv
                    os.path.join(utils.fs.strip_extension(content_path), filename): exp_filesize,
                }
            else:
                # Original release is multiple files (e.g. Extras or Bonus)
                filename = utils.fs.basename(content_path)
                exp_filesizes = {content_path: get_release_filesize(filename)}

        # Compare expected file sizes to actual file sizes
        for filepath, exp_size in exp_filesizes.items():
            filename = utils.fs.basename(filepath)
            actual_size = utils.fs.file_size(filepath)
            _log.debug('Checking file size: %s: %r ?= %r', filename, actual_size, exp_size)
            if exp_size is None:
                _log.debug('No info: %s', filename)
                exceptions.append(errors.SceneMissingInfoError(filename))
            elif actual_size is not None:
                if actual_size != exp_size:
                    _log.debug('Wrong size: %s', filename)
                    exceptions.append(
                        errors.SceneFileSizeError(
                            filename=filename,
                            original_size=exp_size,
                            existing_size=actual_size,
                        )
                    )
                else:
                    _log.debug('Correct size: %s', filepath)
            else:
                _log.debug('No such file: %s', filepath)

        return tuple(e for e in exceptions if e)

    async def verify_release(self, content_path, release_name=None):
        """
        Find matching scene releases and apply :meth:`verify_release_name`
        and :meth:`verify_release_files`

        :param content_path: Path to release file or directory
        :param release_name: Known exact release name or `None` to
            :meth:`search` for `content_path`

        :return: :class:`~.types.SceneCheckResult` enum from
            :meth:`is_scene_release` and sequence of
            :class:`~.errors.SceneError` exceptions from
            :meth:`verify_release_name` and :meth:`verify_release_files`
        """
        # If we know the exact release name, this is easy.
        if release_name:
            return await self._verify_release(content_path, release_name)

        # Find possible `release_name` values. For season packs that were released
        # as single episodes, this will get us a sequence of episode release names.
        existing_release_names = await self.search(content_path)
        if not existing_release_names:
            return SceneCheckResult.false, ()

        # Maybe `content_path` was released by scene as it is (as file or directory)
        for existing_release_name in existing_release_names:
            is_scene_release, exceptions = await self._verify_release(content_path, existing_release_name)
            if is_scene_release and not exceptions:
                return SceneCheckResult.true, ()

        # Maybe `content_path` is a directory (season pack) and scene released
        # single files (episodes).
        return await self._verify_release_per_file(content_path)

    async def _verify_release(self, content_path, release_name):
        _log.debug('Verifying %r against release: %r', content_path, release_name)

        # Stop other checks if this is not a scene release
        is_scene = await self.is_scene_release(release_name)
        if is_scene in (SceneCheckResult.false, SceneCheckResult.unknown):
            return is_scene, ()

        # Combined exceptions from verify_release_name() and verify_release_files()
        exceptions = []

        # verify_release_name() can only produce one exception, so it is raised
        try:
            await self.verify_release_name(content_path, release_name)
        except errors.SceneError as e:
            exceptions.append(e)

        # verify_release_files() can produce multiple exceptions, so it returns them
        exceptions.extend(await self.verify_release_files(content_path, release_name))
        return is_scene, tuple(exceptions)

    async def _verify_release_per_file(self, content_path):
        _log.debug('Verifying each file beneath %r', content_path)
        is_scene_releases = []
        combined_exceptions = collections.defaultdict(lambda: [])
        filepaths = utils.fs.file_list(content_path, extensions=constants.VIDEO_FILE_EXTENSIONS)
        for filepath in filepaths:
            existing_release_names = await self.search(filepath)
            _log.debug('Search results for %r: %r', filepath, existing_release_names)

            # If there are no search results, default to "not a scene release"
            is_scene_release = SceneCheckResult.false

            # Match each existing_release_name against filepath
            for existing_release_name in existing_release_names:
                is_scene_release, exceptions = await self._verify_release(filepath, existing_release_name)
                _log.debug('Verified %r against %r: %r, %r',
                           filepath, existing_release_name, is_scene_release, exceptions)
                if is_scene_release and not exceptions:
                    # Match found, don't check other existing_release_names
                    break
                elif is_scene_release:
                    # Remember exceptions per file (makes debugging easier)
                    combined_exceptions[filepath].extend(exceptions)

            # Remember the SceneCheckResult when the for loop ended. True if we
            # found a scene release at any point, other it's the value of the last
            # existing_release_name.
            is_scene_releases.append(is_scene_release)

        # Collapse `is_scene_releases` into a single value
        if is_scene_releases and all(isr is SceneCheckResult.true for isr in is_scene_releases):
            _log.debug('All files are scene releases')
            is_scene_release = SceneCheckResult.true
        elif is_scene_releases and all(isr is SceneCheckResult.false for isr in is_scene_releases):
            _log.debug('All files are non-scene releases')
            is_scene_release = SceneCheckResult.false
        else:
            _log.debug('Uncertain scene status: %r', is_scene_releases)
            is_scene_release = SceneCheckResult.unknown

        return is_scene_release, tuple(
            exception
            for exceptions in combined_exceptions.values()
            for exception in exceptions
        )



import os
from types import SimpleNamespace
from unittest.mock import AsyncMock, Mock, PropertyMock, call

import pytest

from upsies import errors
from upsies.jobs.poster import PosterJob


def test_cache_id():
    job = PosterJob(webdb=None, webdb_id=None, autostart=False)
    assert job.cache_id is None


@pytest.mark.parametrize(
    argnames='init_kwargs, start_kwargs, exp_work_kwargs',
    argvalues=(
        (
            {'webdb': None, 'webdb_id': None},
            {'webdb': 'start webdb', 'webdb_id': 'start id'},
            {'webdb': 'start webdb', 'webdb_id': 'start id', 'season': None},
        ),
        (
            {'webdb': 'init webdb', 'webdb_id': 'init id'},
            {'webdb': 'start webdb', 'webdb_id': 'start id'},
            {'webdb': 'start webdb', 'webdb_id': 'start id', 'season': None},
        ),
        (
            {'webdb': 'init webdb', 'webdb_id': 'init id', 'season': 'init season'},
            {'webdb': 'start webdb', 'webdb_id': 'start id'},
            {'webdb': 'start webdb', 'webdb_id': 'start id', 'season': 'init season'},
        ),
        (
            {'webdb': 'init webdb', 'webdb_id': 'init id', 'season': 'init season'},
            {'webdb': 'start webdb', 'webdb_id': None, 'season': 'start season'},
            {'webdb': 'start webdb', 'webdb_id': 'init id', 'season': 'start season'},
        ),
        (
            {'webdb': 'init webdb', 'webdb_id': 'init id', 'season': None},
            {'webdb': None, 'webdb_id': 'start id', 'season': 'start season'},
            {'webdb': 'init webdb', 'webdb_id': 'start id', 'season': 'start season'},
        ),
    ),
    ids=lambda v: repr(v),
)
def test_start(init_kwargs, start_kwargs, exp_work_kwargs, mocker):
    mocks = Mock()
    mocker.patch('upsies.jobs.base.JobBase.start', mocks.start)
    mocker.patch('upsies.jobs.poster.PosterJob.attach_task', mocks.attach_task)
    mocker.patch('upsies.jobs.poster.PosterJob._work', mocks._work)

    job = PosterJob(autostart=False, **init_kwargs)

    assert mocks.mock_calls == []
    job.start(**start_kwargs)
    assert mocks.mock_calls == [
        call.start(),
        call._work(**exp_work_kwargs),
        call.attach_task(coro=mocks._work.return_value, finish_when_done=True),
    ]


@pytest.mark.parametrize(
    argnames=(
        'webdb, webdb_id, season, '
        '_get_poster_filepath, _get_poster_url, _download_poster, '
        '_resize_poster, _imghost, _upload_poster, exp_calls'
    ),
    argvalues=(
        (
            'mock webdb',
            'tt123',
            'S03',
            Mock(return_value='downloaded/poster.jpg'),
            AsyncMock(return_value=None),
            AsyncMock(return_value=None),
            Mock(return_value=None),
            None,
            AsyncMock(return_value=None),
            [
                call._get_poster_url('mock webdb', 'tt123', 'S03'),
            ],
        ),
        (
            'mock webdb',
            'tt123',
            'S03',
            Mock(return_value='downloaded/poster.jpg'),
            AsyncMock(return_value='http://original/url.jpg'),
            AsyncMock(return_value=None),
            Mock(return_value=None),
            None,
            AsyncMock(return_value=None),
            [
                call._get_poster_url('mock webdb', 'tt123', 'S03'),
                call._get_poster_filepath('mock webdb', 'tt123', 'S03', 'http://original/url.jpg'),
                call._download_poster('http://original/url.jpg', 'downloaded/poster.jpg'),
            ],
        ),
        (
            'mock webdb',
            'tt123',
            'S03',
            Mock(return_value='downloaded/poster.jpg'),
            AsyncMock(return_value='http://original/url.jpg'),
            AsyncMock(return_value='path/to/original.jpg'),
            Mock(return_value=None),
            None,
            AsyncMock(return_value=None),
            [
                call._get_poster_url('mock webdb', 'tt123', 'S03'),
                call._get_poster_filepath('mock webdb', 'tt123', 'S03', 'http://original/url.jpg'),
                call._download_poster('http://original/url.jpg', 'downloaded/poster.jpg'),
                call._resize_poster('path/to/original.jpg'),
            ],
        ),
        (
            'mock webdb',
            'tt123',
            'S03',
            Mock(return_value='downloaded/poster.jpg'),
            AsyncMock(return_value='http://original/url.jpg'),
            AsyncMock(return_value='path/to/original.jpg'),
            Mock(return_value='path/to/resized.jpg'),
            None,
            AsyncMock(return_value=None),
            [
                call._get_poster_url('mock webdb', 'tt123', 'S03'),
                call._get_poster_filepath('mock webdb', 'tt123', 'S03', 'http://original/url.jpg'),
                call._download_poster('http://original/url.jpg', 'downloaded/poster.jpg'),
                call._resize_poster('path/to/original.jpg'),
                call.send('path/to/resized.jpg'),
            ],
        ),
        (
            'mock webdb',
            'tt123',
            'S03',
            Mock(return_value='downloaded/poster.jpg'),
            AsyncMock(return_value='http://original/url.jpg'),
            AsyncMock(return_value='path/to/original.jpg'),
            Mock(return_value='path/to/resized.jpg'),
            'mock imghost',
            AsyncMock(return_value='http://resized/url.jpg'),
            [
                call._get_poster_url('mock webdb', 'tt123', 'S03'),
                call._get_poster_filepath('mock webdb', 'tt123', 'S03', 'http://original/url.jpg'),
                call._download_poster('http://original/url.jpg', 'downloaded/poster.jpg'),
                call._resize_poster('path/to/original.jpg'),
                call._upload_poster('path/to/resized.jpg'),
                call.send('http://resized/url.jpg'),
            ],
        ),
    ),
)
@pytest.mark.asyncio
async def test_work(webdb, webdb_id, season,
                    _get_poster_filepath, _get_poster_url, _download_poster, _resize_poster, _imghost, _upload_poster,
                    exp_calls, mocker):
    job = PosterJob(autostart=False, webdb=None, webdb_id=None)

    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(job, '_get_poster_filepath', _get_poster_filepath), '_get_poster_filepath')
    mocks.attach_mock(mocker.patch.object(job, '_get_poster_url', _get_poster_url), '_get_poster_url')
    mocks.attach_mock(mocker.patch.object(job, '_download_poster', _download_poster), '_download_poster')
    mocks.attach_mock(mocker.patch.object(job, '_resize_poster', _resize_poster), '_resize_poster')
    job._imghost = _imghost
    mocks.attach_mock(mocker.patch.object(job, '_upload_poster', _upload_poster), '_upload_poster')
    mocks.attach_mock(mocker.patch.object(job, 'send'), 'send')

    await job._work(webdb, webdb_id, season)
    assert mocks.mock_calls == exp_calls


@pytest.mark.parametrize(
    argnames='webdb, webdb_id, season, exp_mock_calls, exp_return_value',
    argvalues=(
        (
            Mock(poster_url=AsyncMock(side_effect=errors.RequestError('No poster available'))),
            'tt123',
            None,
            [
                call.emit('finding_poster', 'tt123'),
                call.poster_url('tt123', season=None),
                call.error('Failed to find poster: No poster available'),
            ],
            None,
        ),
        (
            Mock(poster_url=AsyncMock(return_value='')),
            'tt123',
            'S06',
            [
                call.emit('finding_poster', 'tt123'),
                call.poster_url('tt123', season='S06'),
                call.error('Failed to find poster: There is no poster'),
            ],
            None,
        ),
        (
            Mock(poster_url=AsyncMock(return_value='http://poster/url.jpg')),
            'tt123',
            'S09',
            [
                call.emit('finding_poster', 'tt123'),
                call.poster_url('tt123', season='S09'),
                call.emit('found_poster', 'http://poster/url.jpg'),
            ],
            'http://poster/url.jpg',
        ),
    ),
)
@pytest.mark.asyncio
async def test_get_poster_url(webdb, webdb_id, season, exp_mock_calls, exp_return_value, mocker):
    job = PosterJob(webdb=None, webdb_id=None, autostart=False)

    mocks = Mock()
    mocks.attach_mock(webdb.poster_url, 'poster_url')
    mocks.attach_mock(mocker.patch.object(job, 'error'), 'error')
    mocks.attach_mock(mocker.patch.object(job.signal, 'emit'), 'emit')

    return_value = await job._get_poster_url(webdb, webdb_id, season)
    assert mocks.mock_calls == exp_mock_calls
    assert return_value == exp_return_value


@pytest.mark.parametrize(
    argnames='webdb, webdb_id, season, url, exp_filename',
    argvalues=(
        (
            SimpleNamespace(name='mydb'),
            'tt123',
            None,
            'http://foo/bar',
            'mydb=tt123',
        ),
        (
            SimpleNamespace(name='mydb'),
            'tt123',
            '3',
            'http://foo/bar',
            'mydb=tt123.S03',
        ),
        (
            SimpleNamespace(name='mydb'),
            'tt123',
            '3',
            'http://foo/bar.jpg',
            'mydb=tt123.S03.jpg',
        ),
        (
            SimpleNamespace(name='mydb'),
            'tt123',
            None,
            'http://foo/bar.png',
            'mydb=tt123.png',
        ),
        (
            SimpleNamespace(name='mydb'),
            f'movie{os.sep}123',
            None,
            'http://foo/bar.png',
            'mydb=movie_123.png',
        ),
    ),
)
async def test_get_poster_filepath(webdb, webdb_id, season, url, exp_filename, mocker):
    job = PosterJob(webdb=None, webdb_id=None, autostart=False)

    return_value = job._get_poster_filepath(webdb, webdb_id, season, url)
    assert return_value == os.path.join(job.cache_directory, exp_filename)


@pytest.mark.parametrize(
    argnames='url, filepath, download, exp_mock_calls, exp_return_value',
    argvalues=(
        (
            'http://localhost.org:123/img.jpg',
            'path/to/downloaded/img.jpg',
            AsyncMock(return_value='path/to/sanitized/img.jpg'),
            [
                call.emit('downloading_poster', 'http://localhost.org:123/img.jpg'),
                call.download('http://localhost.org:123/img.jpg', 'path/to/downloaded/img.jpg', cache=True),
                call.emit('downloaded_poster', 'path/to/sanitized/img.jpg'),
            ],
            'path/to/sanitized/img.jpg',
        ),
        (
            'http://localhost.org:123/img.jpg',
            'path/to/downloaded/img.jpg',
            AsyncMock(side_effect=errors.RequestError('Teh internet has broked!')),
            [
                call.emit('downloading_poster', 'http://localhost.org:123/img.jpg'),
                call.download('http://localhost.org:123/img.jpg', 'path/to/downloaded/img.jpg', cache=True),
                call.error('Failed to download poster: Teh internet has broked!'),
            ],
            None,
        ),
    ),
)
@pytest.mark.asyncio
async def test_download_poster(url, filepath, download, exp_mock_calls, exp_return_value, mocker):
    job = PosterJob(autostart=False, webdb=None, webdb_id=None)

    mocks = Mock()
    mocks.attach_mock(mocker.patch('upsies.utils.http.download', download), 'download')
    mocks.attach_mock(mocker.patch.object(job, 'error'), 'error')
    mocks.attach_mock(mocker.patch.object(job.signal, 'emit'), 'emit')

    return_value = await job._download_poster(url, filepath)
    assert mocks.mock_calls == exp_mock_calls
    assert return_value == exp_return_value


@pytest.mark.parametrize(
    argnames='filepath, home_directory, width, height, resize, exp_mock_calls, exp_return_value',
    argvalues=(
        (
            'path/to/original.jpg',
            'path/to/home',
            123,
            456,
            Mock(side_effect=errors.ImageResizeError('Your image is stooopid!')),
            [
                call.emit('resizing_poster', 'path/to/original.jpg'),
                call.resize(
                    'path/to/original.jpg',
                    target_directory='path/to/home',
                    target_filename='original.123x456.jpg',
                    width=123,
                    height=456,
                ),
                call.error('Failed to resize poster: Your image is stooopid!'),
            ],
            None,
        ),
        (
            'path/to/original.jpg',
            'path/to/home',
            123,
            456,
            Mock(return_value='path/to/resized.jpg'),
            [
                call.emit('resizing_poster', 'path/to/original.jpg'),
                call.resize(
                    'path/to/original.jpg',
                    target_directory='path/to/home',
                    target_filename='original.123x456.jpg',
                    width=123,
                    height=456,
                ),
                call.emit('resized_poster', 'path/to/resized.jpg'),
            ],
            'path/to/resized.jpg',
        ),
    ),
)
def test_resize_poster(filepath, home_directory, width, height, resize, exp_mock_calls, exp_return_value, mocker):
    job = PosterJob(autostart=False, webdb=None, webdb_id=None, width=width, height=height)

    mocks = Mock()
    mocks.attach_mock(mocker.patch('upsies.utils.image.resize', resize), 'resize')
    mocks.attach_mock(mocker.patch.object(job, 'error'), 'error')
    mocks.attach_mock(mocker.patch.object(job.signal, 'emit'), 'emit')
    mocker.patch.object(type(job), 'home_directory', PropertyMock(return_value=home_directory))

    return_value = job._resize_poster(filepath)
    assert mocks.mock_calls == exp_mock_calls
    assert return_value == exp_return_value


@pytest.mark.parametrize(
    argnames='filepath, imghost, exp_mock_calls, exp_return_value',
    argvalues=(
        (
            'path/to/file.jpg',
            Mock(upload=AsyncMock(side_effect=errors.RequestError('Nu-uh!'))),
            [
                call.emit('uploading_poster', 'path/to/file.jpg'),
                call.upload('path/to/file.jpg'),
                call.error('Failed to upload poster: Nu-uh!'),
            ],
            None,
        ),
        (
            'path/to/file.jpg',
            Mock(upload=AsyncMock(return_value='http://localhost:123/uploaded.jpg')),
            [
                call.emit('uploading_poster', 'path/to/file.jpg'),
                call.upload('path/to/file.jpg'),
                call.emit('uploaded_poster', 'http://localhost:123/uploaded.jpg'),
            ],
            'http://localhost:123/uploaded.jpg',
        ),
    ),
)
@pytest.mark.asyncio
async def test_upload_poster(filepath, imghost, exp_mock_calls, exp_return_value, mocker):
    job = PosterJob(autostart=False, webdb=None, webdb_id=None, imghost=imghost)

    mocks = Mock()
    mocks.attach_mock(imghost.upload, 'upload')
    mocks.attach_mock(mocker.patch.object(job, 'error'), 'error')
    mocks.attach_mock(mocker.patch.object(job.signal, 'emit'), 'emit')

    return_value = await job._upload_poster(filepath)
    assert mocks.mock_calls == exp_mock_calls
    assert return_value == exp_return_value

    async def _upload_poster(self, filepath):
        try:
            resized_url = await self._imghost.upload(filepath)
        except errors.RequestError as e:
            self.error(f'Failed to upload poster: {e}')
        else:
            self.signal.emit('uploaded_poster', resized_url)
            return resized_url



"""
Image uploader for ptpimg.me
"""

from ... import __project_name__, constants, errors
from .. import configfiles, fs, html, http
from .base import ImageHostBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class PtpimgImageHost(ImageHostBase):
    """Upload images to ptpimg.me"""

    name = 'ptpimg'

    default_config = {
        'base_url': 'https://ptpimg.me',
        'apikey': configfiles.config_value(
            value='',
            description=(
                f'Run ``{__project_name__} ui {name} -h`` '
                'for instructions on how to get an API key.'
            ),
        ),
    }

    description = (
        'You need an API key to upload images.\n'
        '\n'
        '  1. Create an account: https://ptpimg.me/register.php\n'
        f'  2. Store your API key in {fs.tildify_path(constants.IMGHOSTS_FILEPATH)}:\n'
        '\n'
        f'       $ {__project_name__} set --fetch-ptpimg-apikey EMAIL PASSWORD'
    )

    async def _upload_image(self, image_path):
        response = await http.post(
            url=f'{self.options["base_url"]}/upload.php',
            cache=False,
            headers={
                'referer': f'{self.options["base_url"]}/index.php',
            },
            data={
                'api_key': self.options['apikey'],
            },
            files={
                'file-upload[0]': image_path,
            },
        )
        _log.debug('%s: Response: %r', self.name, response)
        images = response.json()

        try:
            code = images[0]['code']
            ext = images[0]['ext']
            assert code and ext, (code, ext)
        except (IndexError, KeyError, TypeError, AssertionError):
            raise RuntimeError(f'Unexpected response: {images}')
        else:
            image_url = f'{self.options["base_url"]}/{code}.{ext}'
            return image_url

    async def get_apikey(self, email, password):
        """
        Get API key from website

        :param str email: Email address to use for login
        :param str password: Password to use for login

        :raises RequestError: if getting the HTML fails for some reason

        :return: API key
        """
        _log.debug('Getting API key for %r', email)
        response = await http.post(
            url=f'{self.options["base_url"]}/login.php',
            cache=False,
            data={
                'email': email,
                'pass': password,
                'login': '',
            },
        )

        try:
            soup = html.parse(response)
            _log.debug('%s: %s', self.name, soup.prettify())

            # Find API key
            input_tag = soup.find('input', id='api_key')
            if input_tag:
                return input_tag['value']

            # Find error message
            error_tag = soup.find(class_='panel-body')
            if error_tag:
                raise errors.RequestError(''.join(error_tag.strings).strip())

            # Default exception
            raise RuntimeError('Failed to find API key')
        finally:
            await http.get(f'{self.options["base_url"]}/logout.php')



import functools

from .. import widgets
from . import JobWidgetBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class ChoiceJobWidget(JobWidgetBase):
    def setup(self):
        self._radiolist = widgets.RadioList(
            options=self.job.options,
            focused=self.job.focused,
            autodetected=self.job.autodetected,
            on_accepted=self._handle_accepted,
        )
        self.job.signal.register('dialog_updated', self._handle_dialog_updated)

    def _handle_dialog_updated(self, job):
        self._radiolist.options = job.options
        self._radiolist.autodetected_index = job.autodetected_index
        self._radiolist.focused_index = job.focused_index
        self.invalidate()

    def _handle_accepted(self, choice):
        self.job.make_choice(choice)

    @functools.cached_property
    def runtime_widget(self):
        return self._radiolist


class TextFieldJobWidget(JobWidgetBase):
    def setup(self):
        self._input_field = widgets.InputField(
            text=self.job.text,
            read_only=self.job.read_only,
            style='class:dialog.text',
            on_accepted=self._on_accepted,
        )
        self.job.signal.register('is_loading', self._handle_is_loading)
        self.job.signal.register('read_only', self._handle_read_only)
        self.job.signal.register('text', self._handle_text)

    def _handle_is_loading(self, is_loading):
        self._input_field.is_loading = is_loading
        self.invalidate()

    def _handle_read_only(self, read_only):
        self._input_field.read_only = read_only
        self.invalidate()

    def _handle_text(self, text):
        self._input_field.text = text
        self.invalidate()

    def _on_accepted(self, buffer):
        self.job.send(buffer.text)

    @functools.cached_property
    def runtime_widget(self):
        return self._input_field



import json
import re
from unittest.mock import AsyncMock, call

import pytest

from upsies import __project_name__, constants, errors
from upsies.utils import fs
from upsies.utils.http import Response
from upsies.utils.imghosts import imgbb


def test_name():
    assert imgbb.ImgbbImageHost.name == 'imgbb'


def test_default_config():
    assert imgbb.ImgbbImageHost.default_config == {
        **imgbb.ImgbbImageHost.default_config_common,
        **{
            'base_url': 'https://api.imgbb.com',
            'apikey': '',
        },
    }


def test_description():
    assert imgbb.ImgbbImageHost.description == (
        'You need an API key to upload images. You can get one by following these steps:\n'
        '\n'
        '  1. Create an account: https://imgbb.com/signup\n'
        '  2. Go to https://api.imgbb.com/ and click on "Get API key".\n'
        f'  3. Store your API key in {fs.tildify_path(constants.IMGHOSTS_FILEPATH)}:\n'
        '\n'
        f'       $ {__project_name__} set imghosts.imgbb.apikey <YOUR API KEY>'
    )


@pytest.mark.asyncio
async def test_upload_image_succeeds(mocker, tmp_path):
    post_mock = mocker.patch('upsies.utils.http.post', AsyncMock(return_value=Response(
        text='{"data":{"url":"https://localhost/path/to/image.png"}}',
        bytes=b'irrelevant',
    )))
    imghost = imgbb.ImgbbImageHost(cache_directory=tmp_path)
    url = await imghost._upload_image('some/image.png')
    assert url == 'https://localhost/path/to/image.png'
    assert post_mock.call_args_list == [call(
        url=imghost.options['base_url'] + '/1/upload',
        cache=False,
        data={
            'key': imghost.options['apikey'],
        },
        files={
            'image': 'some/image.png',
        },
    )]


@pytest.mark.parametrize(
    argnames='response',
    argvalues=(
        '{"data":{"ASDF":"https://some/image.png"}}',
        '{"ASDF":{"url":"https://some/image.png"}}',
    ),
)
@pytest.mark.asyncio
async def test_upload_image_gets_unexpected_response(response, mocker, tmp_path):
    mocker.patch('upsies.utils.http.post', AsyncMock(return_value=Response(
        text=response,
        bytes=b'irrelevant',
    )))
    imghost = imgbb.ImgbbImageHost(cache_directory=tmp_path)
    with pytest.raises(RuntimeError, match=rf'^Unexpected response: {re.escape(response)}$'):
        await imghost._upload_image('some/image.png')

@pytest.mark.asyncio
async def test_upload_image_gets_expected_error(mocker, tmp_path):
    mocker.patch('upsies.utils.http.post', AsyncMock(
        side_effect=errors.RequestError('ignored', text=json.dumps({
            "status_code": 400,
            "error": {
                "message": "Your request sucks",
                "code": 310,
                "context": "CHV\\UploadException"
            },
            "status_txt": "Bad Request"
        }))
    ))
    imghost = imgbb.ImgbbImageHost(cache_directory=tmp_path)
    with pytest.raises(errors.RequestError, match=r'^Bad Request: Your request sucks$'):
        await imghost._upload_image('some/image.png')

@pytest.mark.parametrize(
    argnames='error_text',
    argvalues=(
        [],
        '{invalid json]',
        '{"valid": "json", "with": "missing keys"}',
    ),
)
@pytest.mark.asyncio
async def test_upload_image_gets_unexpected_error(error_text, mocker, tmp_path):
    mocker.patch('upsies.utils.http.post', AsyncMock(
        side_effect=errors.RequestError('ignored', text=error_text),
    ))
    imghost = imgbb.ImgbbImageHost(cache_directory=tmp_path)
    with pytest.raises(errors.RequestError, match=rf'^Upload failed: {re.escape(str(error_text))}$'):
        await imghost._upload_image('some/image.png')



import itertools
import queue
import re
from unittest.mock import DEFAULT, AsyncMock, Mock, PropertyMock, call

import pytest

from upsies import errors
from upsies.jobs import screenshots
from upsies.utils.daemon import MsgType


@pytest.fixture
def job(tmp_path, mocker):
    DaemonProcess_mock = Mock(
        return_value=Mock(
            join=AsyncMock(),
        ),
    )
    mocker.patch('upsies.utils.daemon.DaemonProcess', DaemonProcess_mock)
    return screenshots.ScreenshotsJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        ignore_cache=False,
        content_path='some/path',
        exclude_files=('exclude', 'these', 'files'),
        timestamps=(120,),
        count=2,
    )


@pytest.mark.asyncio  # Ensure aioloop exists
async def test_cache_id(tmp_path):
    job = screenshots.ScreenshotsJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        content_path='some/path',
    )
    assert job.cache_id is None


@pytest.mark.parametrize('from_all_videos', (True, False), ids=['from all videos', 'from first video'])
@pytest.mark.parametrize('optimize', (True, False), ids=['optimize', 'no optimize'])
def test_ScreenshotsJob_initialize(from_all_videos, optimize, tmp_path):
    job = screenshots.ScreenshotsJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        ignore_cache=False,
        content_path='some/path',
        exclude_files=['foo', 'bar', 'baz'],
        timestamps=(120,),
        count=2,
        from_all_videos=from_all_videos,
        optimize=optimize,
    )
    assert job._content_path == 'some/path'
    assert job._exclude_files == ['foo', 'bar', 'baz']
    assert job._screenshots_created == 0
    assert job._screenshots_total == -1
    assert job._timestamps == (120,)
    assert job._count == 2
    assert job._from_all_videos is from_all_videos
    assert job._optimize is optimize
    assert job._screenshots_process is None
    assert job._optimize_process is None
    assert 'screenshots_total' in job.signal.signals

    assert job.output == ()
    assert job.errors == ()
    assert not job.is_finished
    assert job.exit_code is None


@pytest.mark.parametrize(
    argnames='optimize, exp_calls',
    argvalues=(
        (None, [call._execute_screenshots_process()]),
        ('none', [call._execute_screenshots_process()]),
        (0, [call._execute_screenshots_process(), call._execute_optimize_process()]),
        (1, [call._execute_screenshots_process(), call._execute_optimize_process()]),
        ('0', [call._execute_screenshots_process(), call._execute_optimize_process()]),
        ('1', [call._execute_screenshots_process(), call._execute_optimize_process()]),
    ),
    ids=lambda v: repr(v),
)
def test_ScreenshotsJob_execute(optimize, exp_calls, job, mocker):
    mocks = Mock()
    mocker.patch.object(type(job), '_optimize', PropertyMock(return_value=optimize), create=True)
    mocks.attach_mock(
        mocker.patch.object(job, '_execute_screenshots_process'),
        '_execute_screenshots_process',
    )
    mocks.attach_mock(
        mocker.patch.object(job, '_execute_optimize_process'),
        '_execute_optimize_process',
    )

    job.execute()

    assert mocks.mock_calls == exp_calls


def test_ScreenshotsJob__execute_screenshots_process(job, mocker):
    DaemonProcess_mock = mocker.patch('upsies.utils.daemon.DaemonProcess')
    job._execute_screenshots_process()
    assert job._screenshots_process is DaemonProcess_mock.return_value
    assert DaemonProcess_mock.call_args_list == [call(
        name='_screenshots_process',
        target=screenshots._screenshots_process,
        kwargs={
            'content_path': job._content_path,
            'exclude_files': job._exclude_files,
            'timestamps': job._timestamps,
            'count': job._count,
            'from_all_videos': job._from_all_videos,
            'output_dir': job.home_directory,
            'overwrite': job.ignore_cache,
        },
        info_callback=job._handle_info,
        error_callback=job._handle_error,
    )]
    assert job._screenshots_process.start.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='optimize, exp_ignore_dependency_error',
    argvalues=(
        ('default', True),
        ('custom', False),
    ),
    ids=lambda v: repr(v),
)
def test_ScreenshotsJob__execute_optimize_process(optimize, exp_ignore_dependency_error, job, mocker):
    mocker.patch.object(job, '_optimize', optimize)
    DaemonProcess_mock = mocker.patch('upsies.utils.daemon.DaemonProcess')
    job._execute_optimize_process()
    assert job._optimize_process is DaemonProcess_mock.return_value
    assert DaemonProcess_mock.call_args_list == [call(
        name='_optimize_process',
        target=screenshots._optimize_process,
        kwargs={
            'level': job._optimize,
            'ignore_dependency_error': exp_ignore_dependency_error,
        },
        info_callback=job._handle_info,
        error_callback=job._handle_error,
    )]
    assert job._optimize_process.start.call_args_list == [call()]


@pytest.mark.parametrize('is_finished', (False, True))
@pytest.mark.parametrize('all_screenshots_created', (False, True))
@pytest.mark.parametrize(
    argnames='optimize, info, exp_attributes, exp_calls',
    argvalues=(
        (
            False,
            {'screenshots_total': 12},
            {
                '_screenshots_total': 12,
            },
            [
                call.emit('screenshots_total', 12),
            ],
        ),
        (
            False,
            {'screenshot_path': 'path/to/screenshot.png', 'video_path': 'path/to/video.mkv'},
            {
                '_screenshots_by_file': {'path/to/video.mkv': ['path/to/screenshot.png']},
                '_screenshots_created': 1,
            },
            [
                call.job_send('path/to/screenshot.png'),
            ],
        ),
        (
            True,
            {'screenshot_path': 'path/to/screenshot.png', 'video_path': 'path/to/video.mkv'},
            {
                '_screenshots_by_file': {'path/to/video.mkv': ['path/to/screenshot.png']},
                '_screenshots_created': 0,
            },
            [
                call.optimize_process_send(MsgType.info, 'path/to/screenshot.png'),
            ],
        ),
        (
            True,
            {'optimized_screenshot_path': 'path/to/screenshot.png'},
            {
                '_screenshots_by_file': {},
                '_screenshots_created': 1,
            },
            [
                call.job_send('path/to/screenshot.png'),
            ],
        ),
    ),
    ids=lambda v: repr(v),
)
def test_ScreenshotsJob__handle_info(is_finished,
                                     optimize, info, exp_attributes, exp_calls,
                                     all_screenshots_created,
                                     job, mocker):
    mocker.patch('upsies.utils.fs.file_size', return_value=123)
    mocks = Mock()
    mocker.patch.object(type(job), 'is_finished', PropertyMock(return_value=is_finished))
    if optimize:
        mocks.attach_mock(mocker.patch.object(job, '_optimize_process').send, 'optimize_process_send')
    else:
        mocker.patch.object(job, '_optimize_process', None)
    mocks.attach_mock(mocker.patch.object(job, 'send'), 'job_send')
    mocks.attach_mock(mocker.patch.object(job.signal, 'emit'), 'emit')
    mocks.attach_mock(mocker.patch.object(job, 'finish'), 'finish')

    if all_screenshots_created:
        mocker.patch.object(type(job), 'screenshots_created', PropertyMock(return_value=123))
        mocker.patch.object(type(job), 'screenshots_total', PropertyMock(return_value=123))
        exp_calls.append(call.finish())

    job._handle_info(info)

    if is_finished:
        assert mocks.mock_calls == []

    else:
        assert mocks.mock_calls == exp_calls
        for attr, exp_value in exp_attributes.items():
            assert getattr(job, attr) == exp_value


def test_ScreenshotsJob_handle_error_with_exception(job):
    assert job.raised is None
    assert job.errors == ()
    job._handle_error(errors.ScreenshotError('Foo!'))
    assert job.raised == errors.ScreenshotError('Foo!')
    assert job.errors == ()
    assert job.is_finished

def test_ScreenshotsJob_handle_error_with_string(job):
    assert job.errors == ()
    job._handle_error('Foo!')
    assert job.errors == ('Foo!',)
    assert job.is_finished


@pytest.mark.parametrize(
    argnames='screenshots_process, exp_screenshots_process_stopped',
    argvalues=(
        (None, False),
        (Mock(name='screenshots_process'), True),
    ),
)
@pytest.mark.parametrize(
    argnames='optimize_process, exp_optimize_process_stopped',
    argvalues=(
        (None, False),
        (Mock(name='optimize_process'), True),
    ),
)
def test_ScreenshotsJob_finish(screenshots_process, exp_screenshots_process_stopped,
                               optimize_process, exp_optimize_process_stopped,
                               job, mocker):
    job.execute()

    mocker.patch.object(job, '_screenshots_process', screenshots_process)
    mocker.patch.object(job, '_optimize_process', optimize_process)
    mocks = Mock()
    if job._screenshots_process:
        mocks.attach_mock(job._screenshots_process.stop, '_screenshots_process_stop')
    if job._optimize_process:
        mocks.attach_mock(job._optimize_process.stop, '_optimize_process_stop')

    job.finish()

    exp_mock_calls = []
    if exp_screenshots_process_stopped:
        exp_mock_calls.append(call._screenshots_process_stop())
    if exp_optimize_process_stopped:
        exp_mock_calls.append(call._optimize_process_stop())
    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize(
    argnames=('screenshots_total', 'output', 'exp_exit_code'),
    argvalues=(
        (-1, ('a.jpg', 'b.jpg', 'c.jpg'), 0),
        (-1, (), 1),
        (0, (), 0),
        (1, (), 1),
        (3, (), 1),
        (3, ('a.jpg',), 1),
        (3, ('a.jpg', 'b.jpg'), 1),
        (3, ('a.jpg', 'b.jpg', 'c.jpg'), 0),
    ),
)
@pytest.mark.asyncio
async def test_exit_code(screenshots_total, output, exp_exit_code, job):
    assert job.exit_code is None
    for o in output:
        job.send(o)
    job._screenshots_total = screenshots_total
    job.finish()
    await job.wait()
    assert job.is_finished
    assert job.exit_code == exp_exit_code


def test_content_path(job, mocker):
    mock_content_path = Mock()
    job.content_path = mock_content_path
    assert job.content_path is mock_content_path

    mocker.patch.object(type(job), 'is_started', PropertyMock(return_value=True))
    with pytest.raises(RuntimeError, match=r'^Cannot set content_path after job has been started$'):
        job.content_path = Mock()
    assert job.content_path is mock_content_path


def test_exclude_files(job, mocker):
    mock_exclude_files = Mock()
    job.exclude_files = mock_exclude_files
    assert job.exclude_files is mock_exclude_files

    mocker.patch.object(type(job), 'is_started', PropertyMock(return_value=True))
    with pytest.raises(RuntimeError, match=r'^Cannot set exclude_files after job has been started$'):
        job.exclude_files = Mock()
    assert job.exclude_files is mock_exclude_files


def test_from_all_videos(job, mocker):
    mock_from_all_videos = Mock()
    job.from_all_videos = mock_from_all_videos
    assert job.from_all_videos is mock_from_all_videos

    mocker.patch.object(type(job), 'is_started', PropertyMock(return_value=True))
    with pytest.raises(RuntimeError, match=r'^Cannot set from_all_videos after job has been started$'):
        job.from_all_videos = Mock()
    assert job.from_all_videos is mock_from_all_videos


def test_count(job, mocker):
    mock_count = Mock()
    job.count = mock_count
    assert job.count is mock_count

    mocker.patch.object(type(job), 'is_started', PropertyMock(return_value=True))
    with pytest.raises(RuntimeError, match=r'^Cannot set count after job has been started$'):
        job.count = Mock()
    assert job.count is mock_count


def test_timestamps(job, mocker):
    mock_timestamps = Mock()
    job.timestamps = mock_timestamps
    assert job.timestamps is mock_timestamps

    mocker.patch.object(type(job), 'is_started', PropertyMock(return_value=True))
    with pytest.raises(RuntimeError, match=r'^Cannot set timestamps after job has been started$'):
        job.timestamps = Mock()
    assert job.timestamps is mock_timestamps


def test_screenshots_total(job):
    assert job.screenshots_total is job._screenshots_total


def test_screenshots_created(job):
    assert job.screenshots_created is job._screenshots_created


def test_screenshots_by_file(job):
    job._screenshots_by_file = {
        'path/to/foo.mkv': ['screenshot1', 'screenshot2'],
    }

    screenshots_by_file = job.screenshots_by_file
    assert screenshots_by_file == {
        'path/to/foo.mkv': ('screenshot1', 'screenshot2'),
    }

    screenshots_by_file['path/to/foo.mkv'] = 'screenshot4'
    assert job.screenshots_by_file == {
        'path/to/foo.mkv': ('screenshot1', 'screenshot2'),
    }

    screenshots_by_file['path/to/bar.mkv'] = 'screenshot5'
    assert job.screenshots_by_file == {
        'path/to/foo.mkv': ('screenshot1', 'screenshot2'),
    }

    del screenshots_by_file['path/to/foo.mkv']
    assert job.screenshots_by_file == {
        'path/to/foo.mkv': ('screenshot1', 'screenshot2'),
    }


@pytest.fixture
def screenshots_process_mocks(mocker):
    args = {
        'output_queue': Mock(),
        'input_queue': Mock(),
        'content_path': 'path/to/videos',
        'exclude_files': ('files', 'to', 'exclude'),
        'timestamps': ['1:00:00', '2:00:00', '3:00:00'],
        'count': 123,
        'from_all_videos': False,
        'output_dir': 'path/to/my/screenshots',
        'overwrite': 'my overwrite value',
    }
    mocks = Mock()
    mocks.attach_mock(args['output_queue'].put, 'output_queue_put')
    mocks.attach_mock(mocker.patch('upsies.jobs.screenshots._get_video_files'), '_get_video_files')
    mocks.attach_mock(mocker.patch('upsies.jobs.screenshots._map_timestamps'), '_map_timestamps')
    mocks.attach_mock(mocker.patch('upsies.jobs.screenshots._screenshot_video_files'), '_screenshot_video_files')
    return args, mocks

def test__screenshots_process__no_videos_found(screenshots_process_mocks):
    args, mocks = screenshots_process_mocks
    mocks._get_video_files.return_value = []

    return_value = screenshots._screenshots_process(
        args['output_queue'], args['input_queue'],
        content_path=args['content_path'],
        exclude_files=args['exclude_files'],
        timestamps=args['timestamps'],
        count=args['count'],
        from_all_videos=args['from_all_videos'],
        output_dir=args['output_dir'],
        overwrite=args['overwrite'],
    )
    assert return_value is None

    assert mocks.mock_calls == [
        call._get_video_files(
            args['output_queue'],
            content_path=args['content_path'],
            exclude_files=args['exclude_files'],
        ),
        call.output_queue_put((MsgType.error, 'No videos found')),
    ]

@pytest.mark.parametrize('termination_indicated', (False, True), ids=['not terminated', 'terminated'])
@pytest.mark.parametrize('from_all_videos', (False, True), ids=['from_first_video', 'from_all_videos'])
def test__screenshots_process(from_all_videos, termination_indicated, screenshots_process_mocks):
    args, mocks = screenshots_process_mocks
    args['from_all_videos'] = from_all_videos
    mocks._get_video_files.return_value = ('foo.mkv', 'bar.mp4', 'baz.web')
    mocks._map_timestamps.return_value = {
        'foo.mkv': [10, 20, 30],
        'bar.mp4': [10, 20, 29],
        'baz.web': [10, 20, 27],
    }
    if termination_indicated:
        mocks._screenshot_video_files.side_effect = SystemExit('terminated')

    return_value = screenshots._screenshots_process(
        args['output_queue'], args['input_queue'],
        content_path=args['content_path'],
        exclude_files=args['exclude_files'],
        timestamps=args['timestamps'],
        count=args['count'],
        from_all_videos=args['from_all_videos'],
        output_dir=args['output_dir'],
        overwrite=args['overwrite'],
    )
    assert return_value is None

    if from_all_videos:
        exp_video_files = ['foo.mkv', 'bar.mp4', 'baz.web']
    else:
        exp_video_files = ['foo.mkv']

    assert mocks.mock_calls == [
        call._get_video_files(
            args['output_queue'],
            content_path=args['content_path'],
            exclude_files=args['exclude_files'],
        ),
        call._map_timestamps(
            video_files=exp_video_files,
            timestamps=args['timestamps'],
            count=args['count'],
        ),
        call.output_queue_put((MsgType.info, {'screenshots_total': 9})),
        call._screenshot_video_files(
            args['output_queue'], args['input_queue'],
            timestamps_map=mocks._map_timestamps.return_value,
            output_dir=args['output_dir'],
            overwrite=args['overwrite'],
        ),
    ]


def run_optimize_process_test(msgs, exp_calls, output_queue, input_queue, ignore_dependency_error, exp_raises, mocker):
    mocks = Mock()
    mocks.attach_mock(output_queue, 'output_queue')
    mocks.attach_mock(input_queue, 'input_queue')
    mocks.attach_mock(
        mocker.patch('upsies.jobs.screenshots._read_queue_until_empty', side_effect=msgs),
        '_read_queue_until_empty',
    )

    def optimize_mock(file_path, *args, **kwargs):
        if file_path.endswith('missing_dependency'):
            raise errors.DependencyError('Missing dependency')
        elif file_path.endswith('bad'):
            raise errors.ImageOptimizeError(f'bad image: {file_path}')
        else:
            return DEFAULT

    mocks.attach_mock(
        mocker.patch('upsies.utils.image.optimize', side_effect=optimize_mock),
        'optimize',
    )

    if isinstance(exp_raises, Exception):
        with pytest.raises(type(exp_raises), match=rf'^{re.escape(str(exp_raises))}$'):
            screenshots._optimize_process(
                output_queue=output_queue,
                input_queue=input_queue,
                level='my level',
                ignore_dependency_error=ignore_dependency_error,
            )
    else:
        return_value = screenshots._optimize_process(
            output_queue=output_queue,
            input_queue=input_queue,
            level='my level',
            ignore_dependency_error=ignore_dependency_error,
        )
        assert return_value is None

def test__optimize_process_terminates_normally(mocker):
    output_queue, input_queue = Mock(), Mock()
    msgs = (
        (
            (MsgType.info, '/path/to/1.png'),
        ),
        (), (),
        (
            (MsgType.info, '/path/to/2.png'),
            (MsgType.info, '/path/to/3.bad'),
            (MsgType.info, '/path/to/4.png'),
            (MsgType.info, 'missing_dependency'),
        ),
        (), (), (), (), (),
        (
            (MsgType.info, '/path/to/5.png'),
            (MsgType.terminate, 'foo'),
            (MsgType.info, '/path/to/6.png'),
            (MsgType.info, '/path/to/7.png'),
        ),
    )
    exp_calls = [
        call._read_queue_until_empty(input_queue=input_queue),
        call.optimize('/path/to/1.png', level='my level'),
        call.output_queue.put((MsgType.info, {'optimized_screenshot_path': '/path/to/1.png'})),
        call._read_queue_until_empty(input_queue=input_queue),
        call._read_queue_until_empty(input_queue=input_queue),
        call._read_queue_until_empty(input_queue=input_queue),
        call.optimize('/path/to/2.png', level='my level'),
        call.output_queue.put((MsgType.info, {'optimized_screenshot_path': '/path/to/2.png'})),
        call._read_queue_until_empty(input_queue=input_queue),
        call.optimize('/path/to/3.bad', level='my level'),
        call.output_queue.put((MsgType.error, 'bad image: /path/to/3.bad')),
        call._read_queue_until_empty(input_queue=input_queue),
        call.optimize('/path/to/4.png', level='my level'),
        call.output_queue.put((MsgType.info, {'optimized_screenshot_path': '/path/to/4.png'})),
        call._read_queue_until_empty(input_queue=input_queue),
        call._read_queue_until_empty(input_queue=input_queue),
        call._read_queue_until_empty(input_queue=input_queue),
        call._read_queue_until_empty(input_queue=input_queue),
    ]
    ignore_dependency_error = True
    exp_raises = None
    run_optimize_process_test(msgs, exp_calls, output_queue, input_queue, ignore_dependency_error, exp_raises, mocker)

def test__optimize_process_raises_dependency_error(mocker):
    output_queue, input_queue = Mock(), Mock()
    msgs = (
        (
            (MsgType.info, '/path/to/1.png'),
            (MsgType.info, 'missing_dependency'),
            (MsgType.info, '/path/to/2.png'),
        ),
        (),
        (
            (MsgType.info, '/path/to/3.png'),
            (MsgType.terminate, 'foo'),
            (MsgType.info, '/path/to/4.png'),
        ),
    )
    exp_calls = [
        call._read_queue_until_empty(input_queue=input_queue),
        call.optimize('/path/to/1.png', level='my level'),
        call.output_queue.put((MsgType.info, {'optimized_screenshot_path': '/path/to/1.png'})),
        call._read_queue_until_empty(input_queue=input_queue),
        call.optimize('missing_dependency', level='my level'),
        call.output_queue.put((MsgType.info, {'optimized_screenshot_path': '/path/to/2.png'})),
        call._read_queue_until_empty(input_queue=input_queue),
        call.optimize('/path/to/3.bad', level='my level'),
        call.output_queue.put((MsgType.error, 'bad image: /path/to/3.bad')),
        call._read_queue_until_empty(input_queue=input_queue),
        call.optimize('/path/to/4.png', level='my level'),
        call.output_queue.put((MsgType.info, {'optimized_screenshot_path': '/path/to/4.png'})),
        call._read_queue_until_empty(input_queue=input_queue),
        call._read_queue_until_empty(input_queue=input_queue),
        call._read_queue_until_empty(input_queue=input_queue),
        call._read_queue_until_empty(input_queue=input_queue),
    ]
    ignore_dependency_error = False
    exp_raises = errors.DependencyError('Missing dependency')
    run_optimize_process_test(msgs, exp_calls, output_queue, input_queue, ignore_dependency_error, exp_raises, mocker)


@pytest.fixture
def get_video_files_mocks(mocker):
    args = {
        'output_queue': Mock(),
        'content_path': 'path/to/videos',
        'exclude_files': ('files', 'to', 'exclude'),
    }
    mocks = Mock()
    mocks.attach_mock(args['output_queue'].put, 'output_queue_put')
    mocks.attach_mock(mocker.patch('upsies.utils.torrent.filter_files'), 'filter_files')
    mocks.attach_mock(mocker.patch('upsies.utils.video.filter_main_videos'), 'filter_main_videos')
    return args, mocks

@pytest.mark.parametrize(
    argnames='content_path, exclude_files, filter_files_result, filter_main_videos_result, exp_return_value, exp_mock_calls',
    argvalues=(
        (
            'path/to/something',
            ['exclude', 'these'],
            ['foo.mkv', 'bar.mp4', 'baz.webm'],
            ['foo.mkv', 'bar.mp4'],
            ['foo.mkv', 'bar.mp4'],
            [
                call.filter_files('path/to/something', exclude=['exclude', 'these']),
                call.filter_main_videos(['foo.mkv', 'bar.mp4', 'baz.webm']),
            ],
        ),
        (
            'path/to/something',
            ['exclude', 'these'],
            errors.ContentError('Failed to filter files'),
            ['foo.mkv', 'bar.mp4'],
            [],
            [
                call.filter_files('path/to/something', exclude=['exclude', 'these']),
                call.output_queue_put((MsgType.error, 'Failed to filter files')),
            ],
        ),
        (
            'path/to/something',
            ['exclude', 'these'],
            ['foo.mkv', 'bar.mp4', 'baz.webm'],
            errors.ContentError('Failed to filter main videos'),
            [],
            [
                call.filter_files('path/to/something', exclude=['exclude', 'these']),
                call.filter_main_videos(['foo.mkv', 'bar.mp4', 'baz.webm']),
                call.output_queue_put((MsgType.error, 'Failed to filter main videos')),
            ],
        ),
    ),
)
def test__get_video_files(content_path, exclude_files,
                          filter_files_result, filter_main_videos_result,
                          exp_return_value, exp_mock_calls, get_video_files_mocks):
    args, mocks = get_video_files_mocks
    args['content_path'] = 'path/to/something'
    args['exclude_files'] = ('bad', 'files')

    if isinstance(filter_files_result, Exception):
        mocks.filter_files.side_effect = filter_files_result
    else:
        mocks.filter_files.return_value = filter_files_result

    if isinstance(filter_main_videos_result, Exception):
        mocks.filter_main_videos.side_effect = filter_main_videos_result
    else:
        mocks.filter_main_videos.return_value = filter_main_videos_result

    return_value = screenshots._get_video_files(
        args['output_queue'],
        content_path=content_path,
        exclude_files=exclude_files,
    )
    assert return_value == exp_return_value

    assert mocks.mock_calls == exp_mock_calls


def test__screenshot_video_files(mocker):
    output_queue = Mock()
    input_queue = Mock()
    timestamps_map = {
        'path/to/foo.mkv': ['1:00:00', '2:00:00', '3:00:00'],
        'path/to/bar.mkv': ['1:00:01', '2:00:01', '3:00:01'],
        'path/to/baz.mkv': ['1:00:02', '2:00:02', '3:00:02'],
    }
    output_dir = 'path/to/my/screenshots'
    overwrite = 'my overwrite value'

    mocks = Mock()
    mocks.attach_mock(mocker.patch('upsies.jobs.screenshots._maybe_terminate'), '_maybe_terminate')
    mocks.attach_mock(mocker.patch('upsies.jobs.screenshots._screenshot_video_file'), '_screenshot_video_file')

    return_value = screenshots._screenshot_video_files(
        output_queue, input_queue,
        timestamps_map=timestamps_map,
        output_dir=output_dir,
        overwrite=overwrite,
    )
    assert return_value is None

    assert list(itertools.chain.from_iterable(
        [
            call._maybe_terminate(input_queue=input_queue),
            call._screenshot_video_file(
                output_queue, input_queue,
                video_file=video_file,
                timestamp=timestamps,
                output_dir=output_dir,
                overwrite=overwrite,
            )
        ]
        for video_file, timestamps in timestamps_map.items()
    ))


def test__screenshot_video_file(mocker):
    output_queue = Mock()
    input_queue = Mock()
    video_file = 'path/to/foo.mkv'
    timestamps = ['1:00:00', '2:00:00', '3:00:00']
    output_dir = 'path/to/my/screenshots'
    overwrite = 'my overwrite value'

    mocks = Mock()
    mocks.attach_mock(mocker.patch('upsies.jobs.screenshots._maybe_terminate'), '_maybe_terminate')
    mocks.attach_mock(mocker.patch('upsies.jobs.screenshots._make_screenshot'), '_make_screenshot')

    return_value = screenshots._screenshot_video_file(
        output_queue, input_queue,
        video_file=video_file,
        timestamps=timestamps,
        output_dir=output_dir,
        overwrite=overwrite,
    )
    assert return_value is None

    assert mocks.mock_calls == list(itertools.chain.from_iterable(
        [
            call._maybe_terminate(input_queue=input_queue),
            call._make_screenshot(
                output_queue,
                video_file=video_file,
                timestamp=ts,
                output_dir=output_dir,
                overwrite=overwrite,
            )
        ]
        for ts in timestamps
    ))


@pytest.mark.parametrize(
    argnames='video_file, screenshot_result, exp_output_queue_put_calls',
    argvalues=(
        (
            'path/to/foo.mkv',
            'path/to/actual_screenshots.jpg',
            [
                call((MsgType.info, {
                    'video_path': 'path/to/foo.mkv',
                    'screenshot_path': 'path/to/actual_screenshots.jpg',
                })),
            ],
        ),
        (
            'path/to/foo.mkv',
            errors.ScreenshotError('Failed to make screenshot'),
            [
                call((MsgType.error, 'Failed to make screenshot')),
            ],
        ),
    ),
    ids=lambda v: repr(v),
)
def test__make_screenshot(video_file, screenshot_result, exp_output_queue_put_calls, mocker):
    if isinstance(screenshot_result, Exception):
        screenshot_mock = mocker.patch('upsies.utils.image.screenshot', side_effect=screenshot_result)
    else:
        screenshot_mock = mocker.patch('upsies.utils.image.screenshot', return_value=screenshot_result)

    output_queue = Mock()
    timestamp = '01:23:45'
    output_dir = 'my/screenshots'
    overwrite = 'my overwrite value'

    return_value = screenshots._make_screenshot(
        output_queue,
        video_file=video_file,
        timestamp=timestamp,
        output_dir=output_dir,
        overwrite=overwrite,
    )
    assert return_value is None

    assert screenshot_mock.call_args_list == [call(
        video_file=video_file,
        screenshot_file='my/screenshots/foo.mkv.01:23:45.png',
        timestamp=timestamp,
        overwrite=overwrite,
    )]
    assert output_queue.put.call_args_list == exp_output_queue_put_calls


def test__map_timestamps(mocker):
    video_files = [
        'path/to/foo.mkv',
        'path/to/bar.mkv',
        'path/to/baz.mkv',
    ]
    timestamps = [10, 30, '90:00']
    count = 123

    def _validate_timestamps(*args, **kwargs):
        return ', '.join(
            itertools.chain(
                (repr(arg) for arg in args),
                (f'{k}={v!r}' for k, v in kwargs.items()),
            ))

    validate_timestamps_mock = mocker.patch('upsies.jobs.screenshots._validate_timestamps',
                                            side_effect=_validate_timestamps)
    return_value = screenshots._map_timestamps(
        video_files=video_files,
        timestamps=timestamps,
        count=count,
    )
    assert return_value == {
        'path/to/foo.mkv': "video_file='path/to/foo.mkv', timestamps=[10, 30, '90:00'], count=123",
        'path/to/bar.mkv': "video_file='path/to/bar.mkv', timestamps=[10, 30, '90:00'], count=123",
        'path/to/baz.mkv': "video_file='path/to/baz.mkv', timestamps=[10, 30, '90:00'], count=123",
    }
    assert validate_timestamps_mock.call_args_list == [
        call(
            video_file=video_file,
            timestamps=timestamps,
            count=count,
        )
        for video_file in video_files
    ]


@pytest.mark.parametrize(
    argnames='duration, count, timestamps, exp_timestamps',
    argvalues=(
        # Invalid duration
        (-1, 0, [], []),
        (0, 0, [], []),
        (1, 0, [], []),
        (59, 0, [], []),
        (70, 0, [], ['0:00:30', '0:00:45']),
        # Default
        (300, 0, [], ['0:02:25', '0:03:37']),
        # Specific count
        (300, 1, [], ['0:02:25']),
        (300, 2, [], ['0:02:25', '0:03:37']),
        (300, 3, [], ['0:01:12', '0:02:25', '0:03:37']),
        (300, 4, [], ['0:01:12', '0:02:25', '0:03:37', '0:04:13']),
        # Specific timestamps
        (300, 0, [60], ['0:01:00']),
        (300, 0, [60, 60], ['0:01:00']),
        (300, 0, [60, '180'], ['0:01:00', '0:03:00']),
        (300, 0, [60, '180', '0:181'], ['0:01:00', '0:03:00', '0:03:01']),
        (300, 0, [60, '180', '002:01:181'], ['0:01:00', '0:03:00', '0:04:50']),
        (86400, 0, [60, '180', '002:01:181'], ['0:01:00', '0:03:00', '2:04:01']),
        # Specific timestamps and specific count
        (300, 3, [0], ['0:00:00', '0:02:25', '0:03:37']),
        (300, 3, [300], ['0:02:25', '0:03:37', '0:04:50']),
        (300, 3, [0, 301], ['0:00:00', '0:02:25', '0:04:50']),
        (300, 1, [60], ['0:01:00']),
        (300, 2, [60], ['0:01:00', '0:02:55']),
        (300, 3, [60], ['0:01:00', '0:02:55', '0:03:52']),
        (300, 3, [60, 180], ['0:01:00', '0:02:00', '0:03:00']),
        # Timestamps are returned sorted
        (300, 3, [60, 180, 120, 90], ['0:01:00', '0:01:30', '0:02:00', '0:03:00']),
        # Timestamps are deduplicated
        (300, 4, [60, '0:60', 90, '1:00'], ['0:01:00', '0:01:30', '0:03:10', '0:04:00']),
    ),
    ids=lambda v: repr(v),
)
def test__validate_timestamps(duration, count, timestamps, exp_timestamps, mocker):
    duration_mock = mocker.patch('upsies.utils.video.duration', return_value=duration)
    return_value = screenshots._validate_timestamps(
        video_file='foo.mkv',
        timestamps=timestamps,
        count=count,
    )
    assert return_value == exp_timestamps
    assert duration_mock.call_args_list == [call('foo.mkv')]


@pytest.mark.parametrize(
    argnames='get_nowait, exp_exception',
    argvalues=(
        (
            Mock(side_effect=queue.Empty()),
            None,
        ),
        (
            Mock(return_value=(MsgType.info, 'for your consideration')),
            None,
        ),
        (
            Mock(return_value=(MsgType.terminate, 'please stop')),
            SystemExit('Terminated'),
        ),
    ),
    ids=lambda v: repr(v),
)
def test__maybe_terminate(get_nowait, exp_exception):
    input_queue = Mock(get_nowait=get_nowait)
    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            screenshots._maybe_terminate(input_queue=input_queue)
    else:
        return_value = screenshots._maybe_terminate(input_queue=input_queue)
        assert return_value is None


@pytest.mark.parametrize(
    argnames='queued_items, exp_return_value',
    argvalues=(
        (
            [queue.Empty('queue is empty')],
            [],
        ),
        (
            [
                ('typ1', 'msg1'),
                ('typ2', 'msg2'),
                ('typ3', 'msg3'),
                queue.Empty('queue is empty'),
            ],
            [
                ('typ1', 'msg1'),
                ('typ2', 'msg2'),
                ('typ3', 'msg3'),
            ],
        ),
        (
            [
                ('typ1', 'msg1'),
                ('typ2', 'msg2'),
                ('typ3', 'msg3'),
                queue.Empty('queue is empty'),
                ('typ4', 'msg4'),
                ('typ5', 'msg5'),
                ('typ6', 'msg6'),
            ],
            [
                ('typ1', 'msg1'),
                ('typ2', 'msg2'),
                ('typ3', 'msg3'),
            ],
        ),
    ),
    ids=lambda v: repr(v),
)
def test__read_queue_until_empty(queued_items, exp_return_value):
    input_queue = Mock(get=Mock(side_effect=queued_items))
    return_value = screenshots._read_queue_until_empty(input_queue=input_queue)
    assert return_value == exp_return_value

    exp_get_call_count = 0
    for item in queued_items:
        exp_get_call_count += 1
        if isinstance(item, queue.Empty):
            break

    assert input_queue.get.call_args_list == [
        call(timeout=0.01)
        for _ in range(exp_get_call_count)
    ]



import copy
import os
import re
from unittest.mock import AsyncMock, Mock, PropertyMock, call, patch

import pytest

from upsies import __project_name__, errors
from upsies.utils import imghosts


def make_TestImageHost(default_config=None, **kwargs):
    # Avoid NameError bug (https://github.com/python/cpython/issues/87546)
    default_config_ = default_config

    class TestImageHost(imghosts.ImageHostBase):
        name = 'imgw00t'
        default_config = default_config_ or {}

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._upload_image_mock = AsyncMock()

        async def _upload_image(self, image_path):
            return await self._upload_image_mock(image_path)

    return TestImageHost(**kwargs)


def test_cache_directory_property(mocker, tmp_path):
    mocker.patch('upsies.constants.DEFAULT_CACHE_DIRECTORY', 'mock/cache/path')
    imghost = make_TestImageHost()
    assert imghost.cache_directory == 'mock/cache/path'
    imghost = make_TestImageHost(cache_directory=tmp_path)
    assert imghost.cache_directory is tmp_path
    imghost.cache_directory = 'path/to/foo'
    assert imghost.cache_directory == 'path/to/foo'


def test_default_config_without_custom_defaults():
    imghost = make_TestImageHost()
    default_config_common = copy.deepcopy(imghost.default_config_common)
    exp_default_config = {**default_config_common, **imghost.default_config}
    assert imghost.default_config == exp_default_config
    exp_options = {**imghost.default_config}
    assert imghost.options == exp_options
    assert imghost.default_config_common == default_config_common

def test_default_config_with_custom_defaults():
    imghost = make_TestImageHost(default_config={'foo': 1, 'bar': 2})
    default_config_common = copy.deepcopy(imghost.default_config_common)
    exp_default_config = {**default_config_common, **{'foo': 1, 'bar': 2}}
    assert imghost.default_config == exp_default_config
    exp_options = exp_default_config
    assert imghost.options == exp_options
    assert imghost.default_config_common == default_config_common

def test_default_config_with_custom_defaults_and_custom_options():
    imghost = make_TestImageHost(default_config={'foo': 1, 'bar': 2}, options={'foo': 1, 'bar': 99})
    default_config_common = copy.deepcopy(imghost.default_config_common)
    exp_default_config = {**default_config_common, **{'foo': 1, 'bar': 2}}
    assert imghost.default_config == exp_default_config
    exp_options = {**imghost.default_config, **{'foo': 1, 'bar': 99}}
    assert imghost.options == exp_options
    assert imghost.default_config_common == default_config_common

def test_default_config_with_overloaded_common_defaults():
    default_config_key = tuple(imghosts.ImageHostBase.default_config_common)[0]
    default_config = {default_config_key: 'foozbar'}
    imghost = make_TestImageHost(default_config=default_config)
    default_config_common = copy.deepcopy(imghost.default_config_common)
    exp_default_config = {**default_config_common, **{default_config_key: 'foozbar'}}
    assert imghost.default_config == exp_default_config
    exp_options = exp_default_config
    assert imghost.options == exp_options
    assert imghost.default_config_common == default_config_common


def test_options_property():
    imghost = make_TestImageHost()
    assert imghost.options == {'thumb_width': 0}
    imghost = make_TestImageHost(default_config={'foo': 1, 'bar': 2})
    assert imghost.options == {'thumb_width': 0, 'foo': 1, 'bar': 2}
    imghost = make_TestImageHost(default_config={'foo': 1, 'bar': 2}, options={'bar': 99})
    assert imghost.options == {'thumb_width': 0, 'foo': 1, 'bar': 99}


def test_description():
    imghost = make_TestImageHost()
    assert imghost.description == ''


@pytest.mark.parametrize(
    argnames='options, exp_exception',
    argvalues=(
        ({}, None),
        ({'apikey': 'd34db33f'}, None),
        ({'apikey': ''}, errors.RequestError(
            'You must configure an API key first. Run '
            f'"{__project_name__} upload-images {{name}} --help" '
            'for more information.'
        )),
    ),
)
@pytest.mark.asyncio
async def test_upload_checks_for_missing_apikey(options, exp_exception, mocker, tmp_path):
    resize_mock = mocker.patch('upsies.utils.image.resize')
    ih = make_TestImageHost(cache_directory=tmp_path, options=options)
    mocker.patch.object(ih, '_get_image_url', AsyncMock())
    if exp_exception is None:
        await ih.upload('foo.png')
        assert resize_mock.call_args_list == []
        assert ih._get_image_url.call_args_list == [call('foo.png', cache=True)]
    else:
        exp_error = str(exp_exception).format(name=ih.name)
        with pytest.raises(type(exp_exception), match=rf'{re.escape(exp_error)}$'):
            await ih.upload('foo.png')
        assert resize_mock.call_args_list == []
        assert ih._get_image_url.call_args_list == []


@pytest.mark.parametrize('cache', (True, False))
@pytest.mark.parametrize('thumb_width', (0, 123))
@pytest.mark.asyncio
async def test_upload_succeeds(thumb_width, cache, mocker, tmp_path):
    resize_mock = mocker.patch('upsies.utils.image.resize', return_value='thumbnail.png')

    ih = make_TestImageHost(cache_directory=tmp_path, options={'thumb_width': thumb_width})
    image_urls = ['https://localhost:123/foo.png']
    if thumb_width:
        image_urls.append('https://localhost:123/foo.thumb.png')
    mocker.patch.object(ih, '_get_image_url', AsyncMock(side_effect=image_urls))

    image = await ih.upload('path/to/foo.png', cache=cache)

    if thumb_width:
        assert resize_mock.call_args_list == [call(
            'path/to/foo.png',
            width=thumb_width,
            target_directory=ih.cache_directory,
        )]
        assert ih._get_image_url.call_args_list == [
            call('path/to/foo.png', cache=cache),
            call(resize_mock.return_value, cache=cache),
        ]
    else:
        assert ih._get_image_url.call_args_list == [
            call('path/to/foo.png', cache=cache),
        ]

    assert isinstance(image, imghosts.UploadedImage)
    assert image == image_urls[0]
    if thumb_width:
        assert image.thumbnail_url == image_urls[1]
    else:
        assert image.thumbnail_url is None

@pytest.mark.parametrize(
    argnames='cache, resize_error, exp_request_error',
    argvalues=(
        (False, errors.ImageResizeError('not an image'), errors.RequestError('not an image')),
        (True, errors.ImageResizeError('not an image'), errors.RequestError('not an image')),
    ),
)
@pytest.mark.asyncio
async def test_upload_catches_ResizeError(cache, resize_error, exp_request_error, mocker, tmp_path):
    resize_mock = mocker.patch('upsies.utils.image.resize', side_effect=resize_error)

    ih = make_TestImageHost(cache_directory=tmp_path, options={'thumb_width': 345})
    image_urls = ['https://localhost:123/foo.png', 'https://localhost:123/foo.thumb.png']
    mocker.patch.object(ih, '_get_image_url', AsyncMock(side_effect=image_urls))

    with pytest.raises(type(exp_request_error), match=rf'^{re.escape(str(exp_request_error))}$'):
        await ih.upload('path/to/foo.png', cache=cache)

    assert resize_mock.call_args_list == [call(
        'path/to/foo.png',
        width=345,
        target_directory=ih.cache_directory,
    )]
    assert ih._get_image_url.call_args_list == [
        call('path/to/foo.png', cache=cache),
    ]


@pytest.mark.asyncio
async def test_get_image_url_from_cache(mocker, tmp_path):
    ih = make_TestImageHost(cache_directory=tmp_path)
    mocker.patch.object(ih, '_upload_image', AsyncMock(return_value='http://localhost:123/uploaded.image.jpg'))
    mocker.patch.object(ih, '_get_url_from_cache', Mock(return_value='http://localhost:123/cached.image.jpg'))
    mocker.patch.object(ih, '_store_url_to_cache')
    url = await ih._get_image_url('path/to/image.jpg', cache=True)
    assert url == 'http://localhost:123/cached.image.jpg'
    assert ih._get_url_from_cache.call_args_list == [call('path/to/image.jpg')]
    assert ih._upload_image.call_args_list == []
    assert ih._store_url_to_cache.call_args_list == []

@pytest.mark.asyncio
async def test_get_image_url_from_upload(mocker, tmp_path):
    ih = make_TestImageHost(cache_directory=tmp_path)
    mocker.patch.object(ih, '_upload_image', AsyncMock(return_value='http://localhost:123/uploaded.image.jpg'))
    mocker.patch.object(ih, '_get_url_from_cache', Mock(return_value='http://localhost:123/cached.image.jpg'))
    mocker.patch.object(ih, '_store_url_to_cache')
    url = await ih._get_image_url('path/to/image.jpg', cache=False)
    assert url == 'http://localhost:123/uploaded.image.jpg'
    assert ih._get_url_from_cache.call_args_list == []
    assert ih._upload_image.call_args_list == [call('path/to/image.jpg')]
    assert ih._store_url_to_cache.call_args_list == [call(
        'path/to/image.jpg',
        'http://localhost:123/uploaded.image.jpg',
    )]

@pytest.mark.asyncio
async def test_get_image_url_prepends_image_path_to_request_error(mocker, tmp_path):
    ih = make_TestImageHost(cache_directory=tmp_path)
    mocker.patch.object(ih, '_upload_image', AsyncMock(side_effect=errors.RequestError('Connection refused')))
    mocker.patch.object(ih, '_get_url_from_cache', Mock(return_value='http://localhost:123/cached.image.jpg'))
    mocker.patch.object(ih, '_store_url_to_cache')
    with pytest.raises(errors.RequestError, match=r'^Connection refused$'):
        await ih._get_image_url('path/to/image.jpg', cache=False)
    assert ih._get_url_from_cache.call_args_list == []
    assert ih._upload_image.call_args_list == [call('path/to/image.jpg')]
    assert ih._store_url_to_cache.call_args_list == []


def test_get_url_from_cache_succeeds(tmp_path):
    ih = make_TestImageHost(cache_directory=tmp_path)
    image_filepath = os.path.join(tmp_path, 'foo.png')
    cache_file = ih._cache_file(image_filepath)
    with open(cache_file, 'w') as f:
        f.write('http://localhost:123/foo.png')
    url = ih._get_url_from_cache(image_path=image_filepath)
    assert url == 'http://localhost:123/foo.png'

def test_get_url_from_cache_with_nonexisting_cache_file(tmp_path):
    ih = make_TestImageHost(cache_directory=tmp_path)
    image_filepath = os.path.join(tmp_path, 'foo.png')
    url = ih._get_url_from_cache(image_path=image_filepath)
    assert url is None

def test_get_url_from_cache_fails_to_read_cache_file(tmp_path):
    ih = make_TestImageHost(cache_directory=tmp_path)
    cache_file = ih._cache_file(os.path.join(tmp_path, 'foo.png'))
    with open(cache_file, 'w') as f:
        f.write('secret')
    os.chmod(cache_file, 0o000)
    try:
        url = ih._get_url_from_cache(image_path=os.path.join(tmp_path, 'foo.png'))
    finally:
        os.chmod(cache_file, 0o600)
    assert url is None


def test_store_url_to_cache_succeeds(mocker, tmp_path):
    mkdir_mock = mocker.patch('upsies.utils.fs.mkdir')
    ih = make_TestImageHost(cache_directory=tmp_path)
    ih._store_url_to_cache(
        image_path=os.path.join(tmp_path, 'foo.png'),
        url='http://localhost:123/image.jpg',
    )
    assert mkdir_mock.call_args_list == [call(str(tmp_path))]
    cache_file = ih._cache_file(os.path.join(tmp_path, 'foo.png'))
    cache_content = open(cache_file, 'r').read()
    assert cache_content == 'http://localhost:123/image.jpg'

def test_store_url_to_cache_fails_to_write(mocker, tmp_path):
    mkdir_mock = mocker.patch('upsies.utils.fs.mkdir')
    ih = make_TestImageHost(cache_directory=tmp_path)
    cache_file = ih._cache_file(os.path.join(tmp_path, 'foo.png'))
    os.chmod(tmp_path, 0o000)
    try:
        with pytest.raises(RuntimeError, match=rf'^Unable to write cache {cache_file}: Permission denied$'):
            ih._store_url_to_cache(
                image_path=os.path.join(tmp_path, 'foo.png'),
                url='http://localhost:123/image.jpg',
            )
        assert not os.path.exists(cache_file)
        assert mkdir_mock.call_args_list == [call(str(tmp_path))]
    finally:
        os.chmod(tmp_path, 0o700)


@pytest.mark.parametrize(
    argnames='cache_dir, exp_cache_dir',
    argvalues=(
        (None, '/tmp/path'),
        ('some/path', 'some/path'),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.parametrize(
    argnames='cache_id',
    argvalues=(
        None,
        0,
        '',
        'image_id',
    ),
    ids=lambda v: str(v),
)
@pytest.mark.parametrize(
    argnames='filesize',
    argvalues=(
        OSError('no such file ignored here'),
        123,
    ),
    ids=lambda v: str(v),
)
@pytest.mark.parametrize('image_dir', (None, 'some/relative/path', '/absolute/path'))
def test_cache_file(image_dir, filesize, cache_id, cache_dir, exp_cache_dir, mocker):
    mocker.patch('upsies.constants.DEFAULT_CACHE_DIRECTORY', exp_cache_dir)
    ih = make_TestImageHost(cache_directory=cache_dir)
    mocker.patch.object(ih, '_get_cache_id_as_string', return_value=cache_id)
    image_name = 'foo.png'

    exp_cache_name = image_name
    if not isinstance(filesize, OSError):
        exp_cache_name += f'.size={filesize}'
    if cache_id:
        exp_cache_name += f'.{cache_id}'
    exp_cache_name += f'.{ih.name}.url'

    if image_dir is None:
        # Image is in cache dir
        image_path = os.path.join(exp_cache_dir, image_name)
    elif os.path.isabs(image_dir):
        # Image has absolute path
        image_path = os.path.join(image_dir, image_name)
        exp_cache_name = os.path.join(os.path.dirname(image_path), exp_cache_name).replace(os.sep, '_')
    else:
        # Image has relative path
        image_path = os.path.join(os.getcwd(), image_dir, image_name)
        exp_cache_name = os.path.join(os.path.dirname(image_path), exp_cache_name).replace(os.sep, '_')

    exp_cache_file = os.path.join(exp_cache_dir, exp_cache_name)
    if isinstance(filesize, OSError):
        patched_getsize = patch('os.path.getsize', side_effect=filesize)
    else:
        patched_getsize = patch('os.path.getsize', return_value=filesize)

    with patched_getsize:
        assert ih._cache_file(image_path) == exp_cache_file


@pytest.mark.parametrize(
    argnames=('cache_id, exp_cache_id'),
    argvalues=(
        (None, None),
        (23, '23'),
        ('asdf', 'asdf'),
        (['foo', 'bar', (1, 2, 3)], 'foo,bar,1,2,3'),
        ({'foo': 'bar', 'this': [23, 42], (1, 2, 3): 'baz'}, 'foo=bar,this=23,42,1,2,3=baz'),
    ),
)
def test_get_cache_id_as_string(cache_id, exp_cache_id, mocker):
    ih = make_TestImageHost()
    mocker.patch.object(type(ih), 'cache_id', PropertyMock(return_value=cache_id))
    cache_id = ih._get_cache_id_as_string()
    assert cache_id == exp_cache_id



import functools

from .. import widgets
from . import JobWidgetBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class ImageHostJobWidget(JobWidgetBase):
    def setup(self):
        self._upload_progress = widgets.ProgressBar()
        self.job.signal.register('output', self.handle_image_url)
        self.job.signal.register('error', lambda _: self.invalidate())
        self.job.signal.register('finished', lambda _: self.invalidate())

    def handle_image_url(self, url):
        if self.job.images_total > 0:
            self._upload_progress.percent = self.job.images_uploaded / self.job.images_total * 100
            self.invalidate()

    @functools.cached_property
    def runtime_widget(self):
        return self._upload_progress



import asyncio
from types import SimpleNamespace
from unittest.mock import AsyncMock, Mock, PropertyMock, call, patch

import pytest
from prompt_toolkit.application import Application
from prompt_toolkit.input import create_pipe_input
from prompt_toolkit.layout.containers import Window
from prompt_toolkit.output import DummyOutput

from upsies.uis.tui.tui import TUI


@pytest.fixture(autouse='module')
def mock_app(mocker):
    app = Application(
        input=create_pipe_input(),
        output=DummyOutput(),
    )
    mocker.patch('upsies.uis.tui.tui.TUI._make_app', Mock(return_value=app))
    mocker.patch('upsies.uis.tui.tui.TUI._jobs_container', Mock(children=[]), create=True)
    mocker.patch('upsies.uis.tui.tui.TUI._layout', create=True)

@pytest.fixture(autouse='module')
def mock_JobWidget(mocker):
    job_widget = Mock(
        __pt_container__=Mock(return_value=(Window())),
        is_interactive=None,
        job=Mock(wait=AsyncMock()),
    )
    mocker.patch('upsies.uis.tui.jobwidgets.JobWidget', Mock(return_value=job_widget))

class MockJob(Mock):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if kwargs.get('name') is not None:
            self.configure_mock(name=kwargs['name'])

        if 'prejobs' not in kwargs:
            self.configure_mock(prejobs=())


def test_add_jobs(mocker):
    jobs = (
        MockJob(name='a', prejobs=(),),
        MockJob(name='b', prejobs=(
            MockJob(name='b.1', prejobs=()),
            MockJob(name='b.2', prejobs=(
                MockJob(name='b.2.x', prejobs=()),
            )),
        )),
        MockJob(name='c', prejobs=(
            MockJob(name='c.1', prejobs=(
                MockJob(name='c.1.x', prejobs=()),
                MockJob(name='c.1.y', prejobs=()),
            )),
            MockJob(name='c.2', prejobs=()),
        )),
    )

    def find_job(name, jobs):
        for job in jobs:
            if job.name == name:
                return job
            prejob = find_job(name, job.prejobs)
            if prejob:
                return prejob

    ui = TUI()
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(ui, '_add_job'), '_add_job')
    mocks.attach_mock(mocker.patch.object(ui, '_update_jobs_container'), '_update_jobs_container')
    mocks.attach_mock(mocker.patch.object(ui, '_connect_jobs'), '_connect_jobs')
    mocks.attach_mock(mocker.patch.object(ui, '_start_jobs'), '_start_jobs')

    ui.add_jobs(*jobs)

    assert mocks.mock_calls == [
        call._add_job(find_job('a', jobs)),
        call._add_job(find_job('b', jobs)),
        call._add_job(find_job('c', jobs)),
        call._update_jobs_container(),
        call._connect_jobs(find_job('a', jobs), find_job('b', jobs), find_job('c', jobs)),
        call._start_jobs(find_job('a', jobs), find_job('b', jobs), find_job('c', jobs)),
    ]


def test_add_job_detects_job_name_duplicate(mocker):
    mocker.patch('upsies.uis.tui.jobwidgets.JobWidget')
    mocker.patch('upsies.uis.tui.tui.to_container')
    ui = TUI()
    ui._add_job(MockJob(name='a'))
    ui._add_job(MockJob(name='b'))
    with pytest.raises(RuntimeError, match=r'^Conflicting job name: b$'):
        ui._add_job(MockJob(name='b'))


def test_add_job_gracefully_ignores_adding_exact_job_duplicate(mocker):
    mocker.patch('upsies.uis.tui.jobwidgets.JobWidget')
    mocker.patch('upsies.uis.tui.tui.to_container')
    ui = TUI()
    ui._add_job(MockJob(name='a'))
    b_job = MockJob(name='b')
    ui._add_job(b_job)
    ui._add_job(MockJob(name='c'))
    ui._add_job(b_job)
    ui._add_job(MockJob(name='d'))
    assert list(ui._jobs) == ['a', 'b', 'c', 'd']


def test_add_job_creates_JobWidget_and_Container(mocker):
    JobWidget_mock = mocker.patch('upsies.uis.tui.jobwidgets.JobWidget')
    to_container_mock = mocker.patch('upsies.uis.tui.tui.to_container')
    ui = TUI()
    job = MockJob(name='a')
    ui._add_job(job)
    assert tuple(ui._jobs) == (job.name,)
    assert ui._jobs[job.name].widget == JobWidget_mock.return_value
    assert ui._jobs[job.name].container == to_container_mock.return_value
    assert JobWidget_mock.call_args_list == [call(job, ui._app)]
    assert to_container_mock.call_args_list == [call(JobWidget_mock.return_value)]


class JobInfo:
    def __init__(self, *, job, widget, container):
        self.job = job
        self.widget = widget
        self.container = container

    def __repr__(self):
        return f'<{type(self).__name__} {self.job.name!r}>'

class Job:
    def __init__(self, name, **kwargs):
        self.name = name
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __repr__(self):
        return f'<{type(self).__name__} {self.name!r}>'

def test_update_jobs_container_only_adds_first_unfinished_job_and_focuses_it_if_no_job_has_errors(mocker):
    ui = TUI()
    mocker.patch.object(ui, '_update_focus')
    ui._jobs = {
        'an': JobInfo(job=Job('a', is_enabled=False, is_started=False, is_finished=False, errors=()), widget=Mock(is_interactive=False), container=Mock(name='aw')),

        'bi': JobInfo(job=Job('b', is_enabled=False, is_started=False, is_finished=False, errors=()), widget=Mock(is_interactive=True), container=Mock(name='bw')),
        'cn': JobInfo(job=Job('c', is_enabled=False, is_started=True, is_finished=False, errors=()), widget=Mock(is_interactive=False), container=Mock(name='cw')),
        'dn': JobInfo(job=Job('d', is_enabled=True, is_started=False, is_finished=False, errors=()), widget=Mock(is_interactive=False), container=Mock(name='dw')),

        'ei': JobInfo(job=Job('e', is_enabled=False, is_started=True, is_finished=False, errors=()), widget=Mock(is_interactive=True), container=Mock(name='ew')),
        'fi': JobInfo(job=Job('f', is_enabled=True, is_started=False, is_finished=False, errors=()), widget=Mock(is_interactive=True), container=Mock(name='fw')),
        'gn': JobInfo(job=Job('g', is_enabled=True, is_started=True, is_finished=False, errors=()), widget=Mock(is_interactive=False), container=Mock(name='gw')),

        'hi': JobInfo(job=Job('h', is_enabled=True, is_started=True, is_finished=False, errors=()), widget=Mock(is_interactive=True), container=Mock(name='hw')),

        'xi': JobInfo(job=Job('x', is_enabled=True, is_started=True, is_finished=False, errors=()), widget=Mock(is_interactive=True), container=Mock(name='xw')),
        'yn': JobInfo(job=Job('y', is_enabled=True, is_started=True, is_finished=False, errors=()), widget=Mock(is_interactive=False), container=Mock(name='yw')),
        'zi': JobInfo(job=Job('z', is_enabled=True, is_started=True, is_finished=False, errors=()), widget=Mock(is_interactive=True), container=Mock(name='zw')),
    }
    ui._layout = Mock()
    jobs_container_id = id(ui._jobs_container)

    def assert_jobs_container(*keys, focused):
        ui._update_jobs_container()
        assert id(ui._jobs_container) == jobs_container_id
        exp_children = [ui._jobs[k].container for k in keys]
        assert ui._jobs_container.children == exp_children
        if focused:
            assert ui._update_focus.call_args_list == [
                call(ui._jobs[f])
                for f in focused
            ]
        else:
            assert ui._update_focus.call_args_list == []
        ui._update_focus.reset_mock()
        print('#######################################################')

    assert_jobs_container('hi', 'gn', 'yn', focused=['hi'])

    ui._jobs['an'].job.is_finished = True
    assert_jobs_container('hi', 'gn', 'yn', focused=['hi'])

    ui._jobs['bi'].job.is_finished = True
    assert_jobs_container('hi', 'gn', 'yn', focused=['hi'])

    ui._jobs['cn'].job.is_finished = True
    assert_jobs_container('hi', 'gn', 'yn', focused=['hi'])

    ui._jobs['dn'].job.is_finished = True
    assert_jobs_container('hi', 'gn', 'yn', focused=['hi'])

    ui._jobs['ei'].job.is_finished = True
    assert_jobs_container('hi', 'gn', 'yn', focused=['hi'])

    ui._jobs['fi'].job.is_finished = True
    assert_jobs_container('hi', 'gn', 'yn', focused=['hi'])

    ui._jobs['gn'].job.is_finished = True
    assert_jobs_container('hi', 'gn', 'yn', focused=['hi'])

    ui._jobs['hi'].job.is_finished = True
    assert_jobs_container('hi', 'xi', 'gn', 'yn', focused=['xi'])

    ui._jobs['xi'].job.errors = ('Error message',)
    assert_jobs_container('hi', 'xi', 'zi', 'gn', 'yn', focused=['xi', 'zi'])


@pytest.mark.parametrize('raised', (None, ValueError('whatever')))
def test_update_focus(raised, mocker):
    ui = TUI()
    mocker.patch.object(ui._layout, 'focus', side_effect=raised)
    jobinfo = Mock()
    ui._update_focus(jobinfo)
    assert ui._layout.focus.call_args_list == [call(jobinfo.container)]


def test_connect_jobs():
    ui = TUI()
    mocks = Mock()
    jobs = [
        mocks.foo,
        mocks.bar,
        mocks.baz,
    ]
    ui._connect_jobs(*jobs)
    assert mocks.mock_calls == [
        call.foo.signal.register('finished', ui._handle_job_finished),
        call.foo.signal.register('refresh_job_list', ui._refresh_jobs),
        call.bar.signal.register('finished', ui._handle_job_finished),
        call.bar.signal.register('refresh_job_list', ui._refresh_jobs),
        call.baz.signal.register('finished', ui._handle_job_finished),
        call.baz.signal.register('refresh_job_list', ui._refresh_jobs),
    ]


@pytest.mark.parametrize(
    argnames='finished_job, enabled_jobs, exp_mock_calls',
    argvalues=(
        pytest.param(
            Mock(exit_code=1),
            [SimpleNamespace(name='foo', is_finished=False)],
            [call._exit()],
            id='Finished job failed',
        ),
        pytest.param(
            Mock(exit_code=0),
            [SimpleNamespace(name='foo', is_finished=False)],
            [call._refresh_jobs()],
            id='Finished job succeeded, only other job is not finished',
        ),
        pytest.param(
            Mock(exit_code=0),
            [SimpleNamespace(name='foo', is_finished=True)],
            [call._exit()],
            id='Finished job succeeded, only other job is finished',
        ),
        pytest.param(
            Mock(exit_code=0),
            [
                SimpleNamespace(name='foo', is_finished=True),
                SimpleNamespace(name='bar', is_finished=True),
                SimpleNamespace(name='baz', is_finished=True),
            ],
            [call._exit()],
            id='Finished job succeeded, all other jobs are finished',
        ),
        pytest.param(
            Mock(exit_code=0),
            [
                SimpleNamespace(name='foo', is_finished=True),
                SimpleNamespace(name='bar', is_finished=False),
                SimpleNamespace(name='baz', is_finished=True),
            ],
            [call._refresh_jobs()],
            id='Finished job succeeded, one other job is not finished',
        ),
    ),
)
def test_handle_job_finished(finished_job, enabled_jobs, exp_mock_calls, mocker):
    ui = TUI()

    mocks = Mock()
    mocker.patch.object(type(ui), '_enabled_jobs', PropertyMock(return_value=[
        Mock(job=job) for job in enabled_jobs
    ]))
    mocks.attach_mock(mocker.patch.object(ui, '_exit'), '_exit')
    mocks.attach_mock(mocker.patch.object(ui, '_refresh_jobs'), '_refresh_jobs')

    ui._handle_job_finished(finished_job)

    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize('app_terminated', (True, False))
def test_refresh_jobs(app_terminated, mocker):
    ui = TUI()

    enabled_jobs = [
        Mock(name='foo'),
        Mock(name='bar'),
        Mock(name='baz'),
    ]

    mocks = Mock()
    mocker.patch.object(type(ui), '_enabled_jobs', PropertyMock(return_value=[
        SimpleNamespace(job=job) for job in enabled_jobs
    ]))
    mocker.patch.object(ui, '_app_terminated', app_terminated)
    mocks.attach_mock(mocker.patch.object(ui, '_start_jobs'), '_start_jobs')
    mocks.attach_mock(mocker.patch.object(ui, '_update_jobs_container'), '_update_jobs_container')
    mocks.attach_mock(mocker.patch.object(ui._app, 'invalidate'), 'invalidate')

    ui._refresh_jobs()

    if app_terminated:
        assert mocks.mock_calls == []
    else:
        assert mocks.mock_calls == [
            call._start_jobs(*enabled_jobs),
            call._update_jobs_container(),
            call.invalidate(),
        ]


def test_start_jobs():
    ui = TUI()
    mocks = Mock()
    jobs = (
        Mock(is_started=False, autostart=False, start=mocks.start_0),
        Mock(is_started=False, autostart=True, start=mocks.start_1),
        Mock(is_started=True, autostart=False, start=mocks.start_2),
        Mock(is_started=True, autostart=True, start=mocks.start_3),
        Mock(is_started=True, autostart=True, start=mocks.start_4),
        Mock(is_started=True, autostart=False, start=mocks.start_5),
        Mock(is_started=False, autostart=True, start=mocks.start_6),
        Mock(is_started=False, autostart=False, start=mocks.start_7),
    )

    ui._start_jobs(*jobs)

    assert mocks.mock_calls == [
        call.start_1(),
        call.start_6(),
    ]


@pytest.mark.asyncio  # Ensure aioloop exists
async def test_run_calls_add_jobs(mocker):
    ui = TUI()
    mocker.patch.object(ui, 'add_jobs')
    mocker.patch.object(ui._app, 'run')
    ui.run(('a', 'b', 'c'))
    assert ui.add_jobs.call_args_list == [call('a', 'b', 'c')]

@pytest.mark.asyncio  # Ensure aioloop exists
async def test_run_runs_application(mocker):
    ui = TUI()
    mocker.patch.object(ui, 'add_jobs')
    mocker.patch.object(ui._app, 'run')
    ui.run(('a', 'b', 'c'))
    assert ui._app.run.call_args_list == [call(set_exception_handler=False)]

@pytest.mark.asyncio  # Ensure aioloop exists
async def test_run_raises_stored_exception(mocker):
    ui = TUI()
    mocker.patch.object(ui, 'add_jobs')
    mocker.patch.object(ui._app, 'run')
    mocker.patch.object(ui, '_get_exception', Mock(return_value=ValueError('foo')))
    with pytest.raises(ValueError, match=r'^foo$'):
        ui.run(('a', 'b', 'c'))

@pytest.mark.asyncio  # Ensure aioloop exists
async def test_run_returns_first_nonzero_job_exit_code(mocker):
    ui = TUI()
    ui._jobs = {
        'a': SimpleNamespace(job=Mock(exit_code=0)),
        'b': SimpleNamespace(job=Mock(exit_code=1)),
        'c': SimpleNamespace(job=Mock(exit_code=2)),
        'd': SimpleNamespace(job=Mock(exit_code=3)),
    }
    mocker.patch.object(ui, 'add_jobs')
    mocker.patch.object(ui._app, 'run')
    mocker.patch.object(ui, '_get_exception', Mock(return_value=None))
    exit_code = ui.run(('a', 'b', 'c'))
    assert exit_code == 1

@pytest.mark.asyncio  # Ensure aioloop exists
async def test_run_returns_zero_if_all_jobs_finished_successfully(mocker):
    ui = TUI()
    ui._jobs = {
        'a': SimpleNamespace(job=Mock(exit_code=0)),
        'b': SimpleNamespace(job=Mock(exit_code=0)),
        'c': SimpleNamespace(job=Mock(exit_code=0)),
        'd': SimpleNamespace(job=Mock(exit_code=0)),
    }
    mocker.patch.object(ui, 'add_jobs')
    mocker.patch.object(ui._app, 'run')
    mocker.patch.object(ui, '_get_exception', Mock(return_value=None))
    exit_code = ui.run(('a', 'b', 'c'))
    assert exit_code == 0


@pytest.mark.asyncio  # Ensure aioloop exists
async def test_exit_does_nothing_if_already_exited(mocker):
    ui = TUI()
    ui._app_terminated = True
    mocker.patch.object(ui, '_terminate_jobs', AsyncMock())
    ui._exit()
    assert ui._terminate_jobs.call_args_list == []
    assert ui._app_terminated is True

@pytest.mark.asyncio  # Ensure aioloop exists
async def test_exit_waits_for_application_to_run(mocker):
    event_loop = asyncio.get_running_loop()
    ui = TUI()
    mocker.patch.object(ui, '_terminate_jobs', AsyncMock())
    mocker.patch.object(type(ui._app), 'is_running', PropertyMock(return_value=False))
    mocker.patch.object(type(ui._app), 'is_done', PropertyMock(return_value=False))
    # If asycnio.call_soon() is still mocked when the test is finished, pytest
    # just hangs. pytest.mark.asyncio seems to depend on it.
    with patch.object(event_loop, 'call_soon'):
        ui._exit()
        assert event_loop.call_soon.call_args_list == [call(ui._exit)]
    assert ui._terminate_jobs.call_args_list == []
    assert ui._app_terminated is False

@pytest.mark.asyncio
async def test_exit_calls_terminate_jobs(mocker):
    ui = TUI()
    mocker.patch.object(type(ui._app), 'is_running', PropertyMock(return_value=True))
    mocker.patch.object(type(ui._app), 'is_done', PropertyMock(return_value=False))
    mocker.patch.object(type(ui._app), 'exit', Mock())
    mocker.patch.object(ui, '_update_jobs_container', Mock())
    ui._exit()
    await asyncio.sleep(0)
    await asyncio.sleep(0)
    assert ui._app_terminated is True
    assert ui._app.exit.call_args_list == [call()]
    assert ui._update_jobs_container.call_args_list == [call()]

@pytest.mark.asyncio
async def test_exit_handles_exception_from_terminate_jobs(mocker):
    ui = TUI()
    mocker.patch.object(type(ui._app), 'is_running', PropertyMock(return_value=True))
    mocker.patch.object(type(ui._app), 'is_done', PropertyMock(return_value=False))
    mocker.patch.object(type(ui._app), 'exit', Mock())
    mocker.patch.object(ui, '_update_jobs_container', Mock())
    mocker.patch.object(ui, '_terminate_jobs', AsyncMock(side_effect=RuntimeError('foo!')))
    ui._exit()
    await asyncio.sleep(0)
    await asyncio.sleep(0)
    await asyncio.sleep(0)
    assert ui._app_terminated is True
    assert ui._app.exit.call_args_list == [call()]
    assert str(ui._exception) == 'foo!'
    assert isinstance(ui._exception, RuntimeError)
    assert ui._update_jobs_container.call_args_list == [call()]


@pytest.mark.asyncio
async def test_terminate_jobs(mocker):
    ui = TUI()
    ui._jobs = {
        'a': SimpleNamespace(job=Mock(is_started=False, is_finished=False, is_enabled=False, wait=AsyncMock())),
        'b': SimpleNamespace(job=Mock(is_started=False, is_finished=False, is_enabled=True, wait=AsyncMock())),
        'c': SimpleNamespace(job=Mock(is_started=False, is_finished=True, is_enabled=False, wait=AsyncMock())),
        'd': SimpleNamespace(job=Mock(is_started=False, is_finished=True, is_enabled=True, wait=AsyncMock())),
        'e': SimpleNamespace(job=Mock(is_started=True, is_finished=False, is_enabled=False, wait=AsyncMock())),
        'f': SimpleNamespace(job=Mock(is_started=True, is_finished=False, is_enabled=True, wait=AsyncMock())),
        'g': SimpleNamespace(job=Mock(is_started=True, is_finished=True, is_enabled=False, wait=AsyncMock())),
        'h': SimpleNamespace(job=Mock(is_started=True, is_finished=True, is_enabled=True, wait=AsyncMock())),
    }
    mocker.patch.object(ui, '_finish_jobs', Mock())
    await ui._terminate_jobs()
    assert ui._finish_jobs.call_args_list == [call([jobinfo.job for jobinfo in ui._jobs.values()])]
    assert ui._jobs['a'].job.wait.call_args_list == []
    assert ui._jobs['b'].job.wait.call_args_list == []
    assert ui._jobs['c'].job.wait.call_args_list == []
    assert ui._jobs['d'].job.wait.call_args_list == []
    assert ui._jobs['e'].job.wait.call_args_list == [call()]
    assert ui._jobs['f'].job.wait.call_args_list == [call()]
    assert ui._jobs['g'].job.wait.call_args_list == []
    assert ui._jobs['h'].job.wait.call_args_list == []


@pytest.mark.asyncio  # Ensure aioloop exists
async def test_finish_jobs():
    ui = TUI()
    jobs = [
        Mock(is_finished=False, is_enabled=False),
        Mock(is_finished=True, is_enabled=False),
        Mock(is_finished=False, is_enabled=True),
        Mock(is_finished=True, is_enabled=True),
    ]
    ui._finish_jobs(jobs)
    assert jobs[0].finish.call_args_list == [call()]
    assert jobs[1].finish.call_args_list == []
    assert jobs[2].finish.call_args_list == [call()]
    assert jobs[3].finish.call_args_list == []


@pytest.mark.asyncio  # Ensure aioloop exists
async def test_get_exception_from_loop_exception_handler():
    ui = TUI()
    ui._exception = ValueError('asdf')
    ui._jobs = {
        'a': SimpleNamespace(job=Mock(raised=ValueError('foo'), is_enabled=False)),
        'b': SimpleNamespace(job=Mock(raised=None, is_enabled=False)),
        'c': SimpleNamespace(job=Mock(raised=ValueError('bar'), is_enabled=True)),
        'd': SimpleNamespace(job=Mock(raised=None, is_enabled=True)),
    }
    exc = ui._get_exception()
    assert isinstance(exc, ValueError)
    assert str(exc) == 'asdf'

@pytest.mark.asyncio  # Ensure aioloop exists
async def test_get_exception_from_first_failed_enabled_job():
    ui = TUI()
    ui._exception = None
    ui._jobs = {
        'a': SimpleNamespace(job=Mock(raised=ValueError('foo'), is_enabled=False)),
        'b': SimpleNamespace(job=Mock(raised=None, is_enabled=False)),
        'c': SimpleNamespace(job=Mock(raised=ValueError('bar'), is_enabled=True)),
        'd': SimpleNamespace(job=Mock(raised=None, is_enabled=True)),
    }
    exc = ui._get_exception()
    assert isinstance(exc, ValueError)
    assert str(exc) == 'bar'

@pytest.mark.asyncio  # Ensure aioloop exists
async def test_get_exception_returns_None_if_no_exception_raised():
    ui = TUI()
    ui._exception = None
    ui._jobs = {
        'a': SimpleNamespace(job=Mock(raised=None, is_enabled=False)),
        'b': SimpleNamespace(job=Mock(raised=None, is_enabled=False)),
        'c': SimpleNamespace(job=Mock(raised=None, is_enabled=True)),
        'd': SimpleNamespace(job=Mock(raised=None, is_enabled=True)),
    }
    assert ui._get_exception() is None



from unittest.mock import Mock, call

import pytest

from upsies.utils import webdbs


def test_webdbs(mocker):
    existing_webdbs = (Mock(), Mock(), Mock())
    submodules_mock = mocker.patch('upsies.utils.webdbs.submodules')
    subclasses_mock = mocker.patch('upsies.utils.webdbs.subclasses', return_value=existing_webdbs)
    assert webdbs.webdbs() == existing_webdbs
    assert submodules_mock.call_args_list == [call('upsies.utils.webdbs')]
    assert subclasses_mock.call_args_list == [call(webdbs.WebDbApiBase, submodules_mock.return_value)]


def test_webdb_returns_WebdbApiBase_instance(mocker):
    existing_webdbs = (Mock(), Mock(), Mock())
    existing_webdbs[0].configure_mock(name='foo')
    existing_webdbs[1].configure_mock(name='bar')
    existing_webdbs[2].configure_mock(name='baz')
    mocker.patch('upsies.utils.webdbs.webdbs', return_value=existing_webdbs)
    assert webdbs.webdb('bar', config={'foo': 'bar'}) is existing_webdbs[1].return_value
    assert existing_webdbs[1].call_args_list == [call(config={'foo': 'bar'})]

def test_webdb_fails_to_find_webdb(mocker):
    existing_webdbs = (Mock(), Mock(), Mock())
    existing_webdbs[0].configure_mock(name='foo')
    existing_webdbs[1].configure_mock(name='bar')
    existing_webdbs[2].configure_mock(name='baz')
    mocker.patch('upsies.utils.webdbs.webdbs', return_value=existing_webdbs)
    with pytest.raises(ValueError, match='^Unsupported web database: bam$'):
        webdbs.webdb('bam')
    for c in existing_webdbs:
        assert c.call_args_list == []


def test_webdb_names(mocker):
    existing_webdbs = (Mock(), Mock(), Mock())
    existing_webdbs[0].configure_mock(name='FOO')
    existing_webdbs[1].configure_mock(name='bar')
    existing_webdbs[2].configure_mock(name='Baz')
    mocker.patch('upsies.utils.webdbs.webdbs', return_value=existing_webdbs)
    assert webdbs.webdb_names() == ['bar', 'Baz', 'FOO']



"""
Dummy tracker for testing and debugging
"""

import asyncio
import functools
import os

from .. import errors, jobs, utils
from . import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


DummyImageHost = utils.types.ImageHost(allowed=('dummy',))


class DummyTrackerConfig(base.TrackerConfigBase):
    defaults = {
        'base_url': 'http://localhost',
        'username': '',
        'password': '',
        'exclude': (),
        'source': 'DMY',
        'image_host': utils.configfiles.config_value(
            value=utils.types.ListOf(
                item_type=DummyImageHost,
                default=DummyImageHost.options,
                separator=',',
            ),
            description=(
                'List of image hosting service names. The first service is normally used '
                + 'with the others as backup if uploading to the first fails.\n'
                + 'Supported services: ' + ', '.join(DummyImageHost.options)
            ),
        ),
    }

    argument_definitions = {
        'submit': {
            ('--skip-category', '-C'): {
                'help': 'Do not ask for category',
                'action': 'store_true',
            },
            ('--screenshots', '--ss'): {
                'help': 'How many screenshots to make',
                'type': int,
                'default': 3,
            },
            ('--delay', '-d'): {
                'help': 'Number of seconds login, upload and logout take each',
                'type': float,
                'default': 1.0,
            },
        },
        'torrent-create': {
            ('--delay', '-d'): {
                'help': 'Number of seconds login and logout take each',
                'type': float,
                'default': 0.0,
            },
        },
    }


class DummyTrackerJobs(base.TrackerJobsBase):

    @functools.cached_property
    def jobs_before_upload(self):
        return (
            self.create_torrent_job,
            self.screenshots_job,
            self.upload_screenshots_job,
            self.poster_job,
            self.mediainfo_job,
            self.tmdb_job,
            self.imdb_job,
            self.release_name_job,
            self.category_job,
            self.scene_check_job,
        )

    @functools.cached_property
    def category_job(self):
        if not self.options['skip_category']:
            return jobs.dialog.ChoiceJob(
                name=self.get_job_name('category'),
                label='Category',
                precondition=self.make_precondition('category_job'),
                options=(
                    (str(typ).capitalize(), typ)
                    for typ in utils.types.ReleaseType if typ
                ),
                autodetected=self.release_name.type,
                **self.common_job_args(),
            )


class DummyTracker(base.TrackerBase):
    name = 'dummy'
    label = 'DuMmY'

    setup_howto_template = (
        'This is just a no-op tracker for testing and demonstration.'
    )

    TrackerJobs = DummyTrackerJobs
    TrackerConfig = DummyTrackerConfig

    async def _login(self):
        _log.debug('%s: Logging in with %r', self.name, self.options)
        await asyncio.sleep(self.options['delay'])
        self._is_logged_in = True

    async def _logout(self):
        _log.debug('%s: Logging out', self.name)
        await asyncio.sleep(self.options['delay'])
        self._is_logged_in = False

    @property
    def is_logged_in(self):
        return getattr(self, '_is_logged_in', False)

    async def get_announce_url(self):
        _log.debug('%s: Getting announce URL', self.name)
        await asyncio.sleep(self.options['delay'])
        return 'http://localhost:123/f1dd15718/announce'

    async def upload(self, tracker_jobs):
        if tracker_jobs.create_torrent_job.output:
            torrent_file = tracker_jobs.create_torrent_job.output[0]
        else:
            raise errors.RequestError('Torrent file was not created.')
        _log.debug('%s: Uploading %s', self.name, torrent_file)
        await asyncio.sleep(self.options['delay'])
        return f'http://localhost/{os.path.basename(torrent_file)}'



from unittest.mock import AsyncMock, Mock, call

import pytest

from upsies import errors
from upsies.utils.imghosts import imgbox


def test_name():
    assert imgbox.ImgboxImageHost.name == 'imgbox'


@pytest.mark.asyncio
async def test_upload_image_handles_success(tmp_path, mocker):
    pyimgbox_upload_mock = mocker.patch('pyimgbox.Gallery.upload', AsyncMock(return_value=Mock(
        success=True,
        image_url='http://foo.url',
        thumbnail_url='http://foo.thumb.url',
        edit_url='http://foo.edit.url',
    )))
    pyimgbox_close_mock = mocker.patch('pyimgbox.Gallery.close', AsyncMock())
    imghost = imgbox.ImgboxImageHost(cache_directory=tmp_path)
    url = await imghost._upload_image('foo.png')
    assert pyimgbox_upload_mock.call_args_list == [call('foo.png')]
    assert pyimgbox_close_mock.call_args_list == [call()]
    assert url == 'http://foo.url'

@pytest.mark.asyncio
async def test_upload_image_handles_error(tmp_path, mocker):
    mocker.patch('pyimgbox.Gallery.upload', AsyncMock(return_value=Mock(
        success=False,
        error='Something went wrong',
    )))
    pyimgbox_close_mock = mocker.patch('pyimgbox.Gallery.close', AsyncMock())
    imghost = imgbox.ImgboxImageHost(cache_directory=tmp_path)
    with pytest.raises(errors.RequestError, match=r'^Something went wrong$'):
        await imghost._upload_image('foo.png')
    assert pyimgbox_close_mock.call_args_list == [call()]

@pytest.mark.asyncio
async def test_upload_image_closes_gallery_on_exception(tmp_path, mocker):
    mocker.patch('pyimgbox.Gallery.upload', AsyncMock(side_effect=RuntimeError('OH MY DOG')))
    pyimgbox_close_mock = mocker.patch('pyimgbox.Gallery.close', AsyncMock())
    imghost = imgbox.ImgboxImageHost(cache_directory=tmp_path)
    with pytest.raises(RuntimeError, match=r'^OH MY DOG$'):
        await imghost._upload_image('foo.png')
    assert pyimgbox_close_mock.call_args_list == [call()]



"""
Custom job that provide the result of a coroutine
"""

import collections

from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class CustomJob(JobBase):
    """Wrapper around coroutine function or awaitable"""

    @property
    def name(self):
        return self._name

    @property
    def label(self):
        return self._label

    def initialize(self, *, name, label, worker, catch=()):
        """
        Set internal state

        :param name: Name for internal use
        :param label: Name for user-facing use
        :param worker: Coroutine function that takes the job instance as a
            positional argument and returns this job's output

            If `worker` returns any iterable that isn't a string, each item is
            passed to :attr:`send`.

            If `worker` returns `None`, it is interpreted as "no output" and the
            job fails. An empty string is valid output and does not mean the job
            failed.
        :param catch: Sequence of :class:`Exception` classes to catch from
            `worker` and pass to :meth:`~.JobBase.error`

        This job finishes when `worker` returns.

        :class:`~.asyncio.CancelledError` from `worker` is ignored. Any other
        exceptions are passed to :meth:`~.JobBase.exception`.
        """
        self._name = str(name)
        self._label = str(label)
        self._worker = worker
        self._expected_exceptions = tuple(catch)
        self._task = None

    def execute(self):
        self._task = self.attach_task(
            self._catch_errors(self._worker(self)),
            callback=self._handle_worker_done,
        )

    async def _catch_errors(self, coro):
        try:
            return await coro
        except self._expected_exceptions as e:
            self.error(e)

    def _handle_worker_done(self, result):
        if result is not None:
            if (
                    not isinstance(result, str)
                    and isinstance(result, collections.abc.Iterable)
            ):
                for result in result:
                    self.send(result)
            else:
                self.send(result)
        self.finish()
        self.signal.emit('refresh_job_list')

    async def wait(self):
        """Wait for `worker`"""
        if self._task:
            await self._task
        await super().wait()

    def finish(self):
        """Cancel `worker` and finish this job"""
        if self._task:
            self._task.cancel()
        super().finish()



"""
Concrete :class:`~.base.TrackerConfigBase` subclass for PTP
"""

import base64

from ...utils import argtypes, configfiles, types
from ..base import TrackerConfigBase

PtpImageHost = types.ImageHost(disallowed=('dummy',))


class PtpTrackerConfig(TrackerConfigBase):
    defaults = {
        'base_url': base64.b64decode('aHR0cHM6Ly9wYXNzdGhlcG9wY29ybi5tZQ==').decode('ascii'),
        'username': '',
        'password': '',
        'announce_url': configfiles.config_value(
            value='',
            description='Your personal announce URL.',
        ),
        'source': 'PTP',
        'image_host': configfiles.config_value(
            value=types.ListOf(
                item_type=PtpImageHost,
                default=('ptpimg', 'imgbox'),
                separator=',',
            ),
            description=(
                'List of image hosting service names. The first service is normally used '
                'with the others as backup if uploading to the first fails.\n'
                + 'Supported services: ' + ', '.join(PtpImageHost.options)
            ),
        ),
        'screenshots_from_movie': configfiles.config_value(
            value=types.Integer(3, min=3, max=10),
            description='How many screenshots to make for single-video uploads.',
        ),
        'screenshots_from_episode': configfiles.config_value(
            value=types.Integer(2, min=2, max=10),
            description='How many screenshots to make per video for multi-video uploads.',
        ),
        'exclude': (
            r'\.(?i:nfo|txt|jpg|jpeg|png|sfv|md5)$',
            r'/(?i:sample|extra|bonus|feature)',
            r'(?i:.*sample\.[a-z]+)$',
        ),
    }

    argument_definitions = {
        'submit': {
            ('--hardcoded-subtitles', '--hs'): {
                'help': 'Video contains hardcoded subtitles',
                'action': 'store_true',
            },
            ('--no-english-subtitles', '--nes'): {
                'help': (
                    'Whether release contains no English audio and no English subtitles.'
                    '\n'
                    'This is autodetected if an audio or subtitle track with an English '
                    'language tag can be found. Subtitle language tags are detected in '
                    '*.idx/sub, VIDEO_TS trees and *.srt/ssa/ass/vtt (by language code '
                    'in the file name, e.g. "Foo.en.srt").'
                ),
                'metavar': 'BOOL',
                'type': argtypes.bool_or_none,
            },
            ('--not-main-movie', '--nmm'): {
                'help': 'Upload ONLY contains extras, Rifftrax, Workprints',
                'action': 'store_true',
            },
            ('--personal-rip', '--pr'): {
                'help': 'Tag as your own encode',
                'action': 'store_true',
            },
            ('--cover-art', '--ca'): {
                'help': (
                    'Path or URL to movie poster\n'
                    'NOTE: This is ignored if a group for this movie already exists.'
                ),
            },
            ('--screenshots', '--ss'): {
                'help': 'How many screenshots to make per video file',
                'type': argtypes.number_of_screenshots(min=1, max=10),
            },
            ('--upload-token', '--ut'): {
                'help': 'Upload token from staff',
            },
            ('--only-description', '--od'): {
                'help': 'Only generate description (do not upload anything)',
                'action': 'store_true',
                'group': 'generate-metadata',
            },
        },
    }



import pytest


@pytest.fixture
def assert_config_list_of_choice():
    def assert_config_list_of_choice(items, exp_items, exp_options, exp_description=None):
        assert items == exp_items
        for item in items:
            assert item.options == tuple(sorted(exp_options))

            cls = type(item)
            for option in item.options:
                assert cls(option) == option
                assert cls(option).options == tuple(sorted(exp_options))

            exp_options_string = ', '.join(sorted(exp_options))
            with pytest.raises(ValueError, match=rf'^Not one of {exp_options_string}: foo$'):
                cls('foo')

        if exp_description is not None:
            assert items.description == exp_description

    return assert_config_list_of_choice


@pytest.fixture
def assert_config_number():
    def assert_config_number(number, value, min, max, description=None):
        assert number == value
        assert number.min == min
        assert number.max == max

        cls = type(number)
        for i in range(min, max + 1):
            assert cls(i) == i
            assert cls(i).min == min
            assert cls(i).max == max

        with pytest.raises(ValueError, match=rf'^Minimum is {min}$'):
            cls(min - 1)

        with pytest.raises(ValueError, match=rf'^Maximum is {max}$'):
            cls(max + 1)

        if description is not None:
            assert number.description == description

    return assert_config_number



import os
import re

import pytest

from upsies import errors
from upsies.utils.predbs import common
from upsies.utils.types import ReleaseType

abbreviated_file_name_test_cases = (
    ('group-thetitle.1080p.bluray.x264.mkv', True),
    ('group-the.title.2016.720p.bluray.x264.mkv', True),
    ('group-thetitle720.mkv', True),
    ('group-the-title-720p.mkv', True),
    ('group-thetitle-720p.mkv', True),
    ('group-ttl-s02e01-720p.mkv', True),
    ('group-the.title.1984.720p.mkv', True),
    ('group-the.title.720p.mkv', True),
    ('group-ttl.720p.mkv', True),
    ('group-title.mkv', True),
    ('title-1080p-group.mkv', True),
    ('group-the_title_x264_bluray.mkv', True),
    ('ttl.720p-group.mkv', True),
    ('GR0up1080pAsDf.mkv', True),
    ('grpthetitle.mkv', True),
    ('title.2017.720p.bluray.x264-group.mkv', False),
)

@pytest.mark.parametrize('path', ('path/to', ''))
@pytest.mark.parametrize(
    argnames='filename, exp_is_abbreviated',
    argvalues=abbreviated_file_name_test_cases,
)
def test_is_abbreviated_filename(path, filename, exp_is_abbreviated):
    filepath = os.path.join(path, filename)
    return_value = common.is_abbreviated_filename(filepath)
    assert return_value is exp_is_abbreviated

@pytest.mark.parametrize('path', ('path/to', ''))
@pytest.mark.parametrize(
    argnames='filename, exp_is_abbreviated',
    argvalues=abbreviated_file_name_test_cases,
)
def test_assert_not_abbreviated_filename(path, filename, exp_is_abbreviated):
    filepath = os.path.join(path, filename)
    exp_msg = f'Abbreviated scene file name is verboten: {filepath}'

    if exp_is_abbreviated:
        with pytest.raises(errors.SceneAbbreviatedFilenameError, match=rf'^{re.escape(exp_msg)}$'):
            common.assert_not_abbreviated_filename(filepath)
    else:
        common.assert_not_abbreviated_filename(filepath)


@pytest.mark.parametrize(
    argnames='source, irrelevant_keys',
    argvalues=(
        ('BluRay', ()),
        ('DVDRip', ('resolution')),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.parametrize(
    argnames='release_info, exclude, exp_keys',
    argvalues=(
        (
            {'type': ReleaseType.movie},
            (),
            ('title', 'year', 'resolution', 'source', 'video_codec', 'group'),
        ),
        (
            {'type': ReleaseType.movie},
            ('year',),
            ('title', 'resolution', 'source', 'video_codec', 'group'),
        ),
        (
            {'type': ReleaseType.season},
            (),
            ('title', 'episodes', 'resolution', 'source', 'video_codec', 'group'),
        ),
        (
            {'type': ReleaseType.season},
            ('resolution',),
            ('title', 'episodes', 'source', 'video_codec', 'group'),
        ),
        (
            {'type': ReleaseType.episode},
            (),
            ('title', 'episodes', 'resolution', 'source', 'video_codec', 'group'),
        ),
        (
            {'type': ReleaseType.episode},
            ('group',),
            ('title', 'episodes', 'resolution', 'source', 'video_codec'),
        ),
        (
            {'type': None},
            (),
            (),
        ),
        (
            {'type': None},
            ('source',),
            (),
        ),
        (
            {'type': 'something weird'},
            (),
            (),
        ),
    ),
    ids=lambda v: str(v),
)
def test_get_needed_keys(release_info, exclude, exp_keys, source, irrelevant_keys):
    release_info['source'] = source
    exp_keys = tuple(k for k in exp_keys if k not in irrelevant_keys)
    needed_keys = common.get_needed_keys(release_info, exclude=exclude)
    assert needed_keys == exp_keys


@pytest.mark.parametrize(
    argnames='release_name, exp_season_pack',
    argvalues=(
        ('X.S01.1080p.BluRay.x264-ASDF', 'X.S01.1080p.BluRay.x264-ASDF'),
        ('X.S01E01.1080p.BluRay.x264-ASDF', 'X.S01.1080p.BluRay.x264-ASDF'),
        ('X.S01E01.Episode.Title.1080p.BluRay.x264-ASDF', 'X.S01.1080p.BluRay.x264-ASDF'),
        ('X.S01E01.Episode.Title.1080i.BluRay.x264-ASDF', 'X.S01.1080i.BluRay.x264-ASDF'),
        ('X.S01E01.Episode.Title.DVDRip.x264-ASDF', 'X.S01.DVDRip.x264-ASDF'),
        ('X.S01E01.Episode.Title.REPACK.720p.BluRay.x264-ASDF', 'X.S01.REPACK.720p.BluRay.x264-ASDF'),
        ('X.S01E01.Episode.Title.PROPER.720p.BluRay.x264-ASDF', 'X.S01.PROPER.720p.BluRay.x264-ASDF'),
        ('X.S01E01.Episode.Title.DUTCH.DVDRip.x264-ASDF', 'X.S01.DUTCH.DVDRip.x264-ASDF'),
        ('X.S01E01.Episode.Title.iNTERNAL.DVDRip.x264-ASDF', 'X.S01.iNTERNAL.DVDRip.x264-ASDF'),
        ('X.S01E01.Episode.Title.REAL.DVDRip.x264-ASDF', 'X.S01.REAL.DVDRip.x264-ASDF'),
        ('X.S01E01.French.DVDRip.x264-ASDF', 'X.S01.French.DVDRip.x264-ASDF'),
        ('X.S01E01.Episode.Title.German.DL.DVDRip.x264-ASDF', 'X.S01.German.DL.DVDRip.x264-ASDF'),
        ('X.S01E01E02.Episode.Title.German.DL.DVDRip.x264-ASDF', 'X.S01.German.DL.DVDRip.x264-ASDF'),
        ('X.S01E03-E04.Episode.Title.German.DL.DVDRip.x264-ASDF', 'X.S01.German.DL.DVDRip.x264-ASDF'),
    ),
)
def test_get_season_pack_name(release_name, exp_season_pack):
    season_pack = common.get_season_pack_name(release_name)
    assert season_pack == exp_season_pack


@pytest.mark.parametrize(
    argnames='directory_name, files, exp_return_value',
    argvalues=(
        pytest.param(
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv',
            (
                'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY.mkv',
            ),
            False,
            id='Movie as file',
        ),

        pytest.param(
            'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY',
            (
                'Dellamorte.Dellamore.1994.1080p.BluRay.x264-LiViDiTY/ly-dellmdm1080p.mkv',
            ),
            False,
            id='Movie in directory',
        ),

        pytest.param(
            'Fawlty.Towers.S01.720p.BluRay.x264-SHORTBREHD',
            (
                'Fawlty.Towers.S01.720p.BluRay.x264-SHORTBREHD/Fawlty.Towers.S01E01.720p.BluRay.x264-SHORTBREHD.mkv',
                'Fawlty.Towers.S01.720p.BluRay.x264-SHORTBREHD/Fawlty.Towers.S01E02.720p.BluRay.x264-SHORTBREHD.mkv',
                'Fawlty.Towers.S01.720p.BluRay.x264-SHORTBREHD/Fawlty.Towers.S01E03.720p.BluRay.x264-SHORTBREHD.mkv',
            ),
            False,
            id='Non-mixed season pack',
        ),

        pytest.param(
            'Drunk.History.S01.720p.HDTV.x264-MiXED',
            (
                'Drunk.History.S01.720p.HDTV.x264-MiXED/drunk.history.s01e01.720p.hdtv.x264-killers.mkv',
                'Drunk.History.S01.720p.HDTV.x264-MiXED/drunk.history.s01e02.720p.hdtv.x264-2hd.mkv',
                'Drunk.History.S01.720p.HDTV.x264-MiXED/drunk.history.s01e03.720p.hdtv.x264.mkv',
                'Drunk.History.S01.720p.HDTV.x264-MiXED/Drunk.History.S01E04.720p.HDTV.x264-EVOLVE.mkv',
            ),
            True,
            id='Season pack with 2 scene groups',
        ),

        pytest.param(
            'Drunk.History.S01.720p.HDTV.x264-MiXED',
            (
                'Drunk.History.S01.720p.HDTV.x264-MiXED/drunk.history.s01e01.720p.hdtv.x264-killers.mkv',
                'Drunk.History.S01.720p.HDTV.x264-MiXED/drunk.history.s01e02.720p.hdtv.x264-2hd.mkv',
                'Drunk.History.S01.720p.HDTV.x264-MiXED/drunk.history.s01e03.720p.hdtv.x264-2hd.mkv',
                'Drunk.History.S01.720p.HDTV.x264-MiXED/Drunk.History.S01E04.720p.HDTV.x264-EVO.mkv',
            ),
            True,
            id='Season pack with 2 scene groups and 1 non-scene group',
        ),

        pytest.param(
            'Drunk.History.S01.720p.HDTV.x264',
            (
                'Drunk.History.S01.720p.HDTV.x264/drunk.history.s01e01.720p.hdtv.x264-killers.mkv',
                'Drunk.History.S01.720p.HDTV.x264/drunk.history.s01e02.720p.hdtv.x264.mkv',
                'Drunk.History.S01.720p.HDTV.x264/Drunk.History.S01E04.720p.HDTV.x264-EVO.mkv',
            ),
            True,
            id='Season pack with 1 scene group and 1 non-scene group and 1 nogroup',
        ),
    ),
)
def test_is_mixed_scene_release(directory_name, files, exp_return_value, tmp_path):
    for filepath in files:
        filepath = tmp_path / filepath
        filepath.parent.mkdir(parents=True, exist_ok=True)
        filepath.write_text(f'this is {filepath}')

    directory_path = str(tmp_path / directory_name)
    return_value = common.is_mixed_season_pack(directory_path)
    assert return_value == exp_return_value



import base64

import pytest

from upsies import utils
from upsies.trackers.ptp import PtpTrackerConfig


@pytest.fixture
def tracker_config():
    return PtpTrackerConfig()


def test_PtpTrackerConfig_defaults(tracker_config):
    assert set(tracker_config) == {
        'base_url',
        'username',
        'password',
        'announce_url',
        'source',
        'randomize_infohash',
        'image_host',
        'screenshots_from_movie',
        'screenshots_from_episode',
        'exclude',

        # Inherited from TrackerConfigBase
        'add_to',
        'copy_to',
    }


def test_PtpTrackerConfig_defaults_base_url(tracker_config):
    assert tracker_config['base_url'] == base64.b64decode('aHR0cHM6Ly9wYXNzdGhlcG9wY29ybi5tZQ==').decode('ascii')


def test_PtpTrackerConfig_defaults_username(tracker_config):
    assert tracker_config['username'] == ''


def test_PtpTrackerConfig_defaults_password(tracker_config):
    assert tracker_config['password'] == ''


def test_PtpTrackerConfig_defaults_announce_url(tracker_config):
    assert tracker_config['announce_url'] == ''
    assert tracker_config['announce_url'].description == (
        'Your personal announce URL.'
    )


def test_PtpTrackerConfig_defaults_source(tracker_config):
    assert tracker_config['source'] == 'PTP'


def test_PtpTrackerConfig_defaults_image_host(tracker_config, assert_config_list_of_choice):
    exp_options = tuple(
        name
        for name in utils.imghosts.imghost_names()
        if name != 'dummy'
    )
    assert_config_list_of_choice(
        items=tracker_config['image_host'],
        exp_items=('ptpimg', 'imgbox'),
        exp_options=exp_options,
        exp_description=(
            'List of image hosting service names. The first service is normally used '
            + 'with the others as backup if uploading to the first fails.\n'
            + 'Supported services: ' + ', '.join(exp_options)
        ),
    )


def test_PtpTrackerConfig_defaults_screenshots_from_movie(tracker_config, assert_config_number):
    assert_config_number(
        number=tracker_config['screenshots_from_movie'],
        value=3,
        min=3,
        max=10,
        description='How many screenshots to make for single-video uploads.',
    )


def test_PtpTrackerConfig_defaults_screenshots_from_episode(tracker_config, assert_config_number):
    assert_config_number(
        number=tracker_config['screenshots_from_episode'],
        value=2,
        min=2,
        max=10,
        description='How many screenshots to make per video for multi-video uploads.',
    )


def test_PtpTrackerConfig_defaults_exclude(tracker_config):
    assert tracker_config['exclude'] == (
        utils.types.RegEx(r'\.(?i:nfo|txt|jpg|jpeg|png|sfv|md5)$'),
        utils.types.RegEx(r'/(?i:sample|extra|bonus|feature)'),
        utils.types.RegEx(r'(?i:.*sample\.[a-z]+)$'),
    )


def test_PtpTrackerConfig_arguments(tracker_config):
    exp_argument_definitions = {
        'submit': {
            ('--hardcoded-subtitles', '--hs'),
            ('--no-english-subtitles', '--nes'),
            ('--not-main-movie', '--nmm'),
            ('--personal-rip', '--pr'),
            ('--cover-art', '--ca'),
            ('--screenshots', '--ss'),
            ('--upload-token', '--ut'),
            ('--only-description', '--od'),
        },
    }
    assert set(tracker_config.argument_definitions) == set(exp_argument_definitions)
    for command in exp_argument_definitions:
        assert set(tracker_config.argument_definitions[command]) == exp_argument_definitions[command]


def test_PtpTrackerConfig_argument_definitions_submit_hardcoded_subtitles(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--hardcoded-subtitles', '--hs')] == {
        'help': 'Video contains hardcoded subtitles',
        'action': 'store_true',
    }


def test_PtpTrackerConfig_argument_definitions_submit_no_english_subtitles(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--no-english-subtitles', '--nes')] == {
        'help': (
            'Whether release contains no English audio and no English subtitles.'
            '\n'
            'This is autodetected if an audio or subtitle track with an English '
            'language tag can be found. Subtitle language tags are detected in '
            '*.idx/sub, VIDEO_TS trees and *.srt/ssa/ass/vtt (by language code '
            'in the file name, e.g. "Foo.en.srt").'
        ),
        'metavar': 'BOOL',
        'type': utils.argtypes.bool_or_none,
    }


def test_PtpTrackerConfig_argument_definitions_submit_not_main_movie(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--not-main-movie', '--nmm')] == {
        'help': 'Upload ONLY contains extras, Rifftrax, Workprints',
        'action': 'store_true',
    }


def test_PtpTrackerConfig_argument_definitions_submit_personal_rip(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--personal-rip', '--pr')] == {
        'help': 'Tag as your own encode',
        'action': 'store_true',
    }


def test_PtpTrackerConfig_argument_definitions_submit_cover_art(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--cover-art', '--ca')] == {
        'help': (
            'Path or URL to movie poster\n'
            'NOTE: This is ignored if a group for this movie already exists.'
        ),
    }


def test_PtpTrackerConfig_argument_definitions_submit_screenshots(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--screenshots', '--ss')] == {
        'help': 'How many screenshots to make per video file',
        'type': utils.argtypes.number_of_screenshots(min=1, max=10),
    }


def test_PtpTrackerConfig_argument_definitions_submit_upload_token(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--upload-token', '--ut')] == {
        'help': 'Upload token from staff',
    }



import pytest

from upsies.utils import image


@pytest.mark.parametrize(
    argnames='os, exp_executable',
    argvalues=(
        ('linux', 'ffmpeg'),
        ('windows', 'ffmpeg.exe'),
        ('plan9', 'ffmpeg'),
    ),
    ids=lambda v: str(v),
)
def test_ffmpeg_executable(os, exp_executable, mocker):
    mocker.patch('upsies.utils.os_family', return_value=os)
    assert image._ffmpeg_executable() == exp_executable
